/*
* Copyright (c) 2018 Xilinx Inc
*
* This file is part of FFmpeg.
*
* FFmpeg is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* FFmpeg is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with FFmpeg; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
*/

#include "libavutil/internal.h"
#include "libavutil/common.h"
#include "libavutil/opt.h"
#include "libavutil/pixdesc.h"
#include "libavutil/imgutils.h"
#include "libavutil/timestamp.h"
#include "libavutil/macros.h"
#include "libavutil/fifo.h"
#include "avcodec.h"
#include "internal.h"
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <memory.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <xma.h>
#include <xrm.h>
#include <pthread.h>
#include "xlnx_lookahead.h"
#include <xvbm.h>
#include <dlfcn.h>
#include "../xmaPropsTOjson.h"

#define SCLEVEL1 2

#define MAX_INPUT_WIDTH     1920
#define MAX_INPUT_HEIGHT    1080
#define MAX_INPUT_PIXELS    (MAX_INPUT_WIDTH * MAX_INPUT_HEIGHT)

#define OFFSET(x) offsetof(mpsoc_vcu_enc_ctx, x)
#define VE AV_OPT_FLAG_VIDEO_PARAM | AV_OPT_FLAG_ENCODING_PARAM

//TODO:comment when not in demo mode
static int global_enc_chan_id= 40;//18; for 32 dec+32 enc config

typedef struct mpsoc_vcu_enc_ctx {
    const AVClass *class;
    XmaEncoderSession *enc_session;
    XmaParameter enc_params[3];
    xrmContext       *xrm_ctx;
    xrmCuListResource  encode_cu_list_res;
    int                encode_res_inuse;
    int ideal_latency;
    XmaFrame frame;
    XmaDataBuffer xma_buffer;
    bool sent_flush;
    int control_rate;
    int64_t max_bitrate;
    int slice_qp;
    int min_qp;
    int max_qp;
    double cpb_size;
    double initial_delay;
    int gop_mode;
    int gdr_mode;
    int b_frames;
    int periodicity_idr;
    int profile;
    int level;
    int tier;
    int num_slices;
    int qp_mode;
    int filler_data;
    int aspect_ratio;
    int dependent_slice;
    int slice_size;
    int scaling_list;
    int entropy_mode;
    int loop_filter;
    int constrained_intra_pred;
    int prefetch_buffer;
    int cores;
    int latency_logging;
    char enc_options[2048];
    AVFifoBuffer *pts_queue;
    int64_t pts_0;
    int64_t pts_1;
    int is_first_outframe;
    const char *lambda_file;
    //LA
    xlnx_lookahead_t la;
    int32_t lookahead_depth;
    AVFifoBuffer *laQ;
    int32_t spatial_aq;
    int32_t temporal_aq;
    int32_t rate_control_mode;
    int32_t spatial_aq_gain;
} mpsoc_vcu_enc_ctx;

int vcu_alloc_ff_packet(mpsoc_vcu_enc_ctx *ctx, AVPacket *pkt);

static const AVOption h264Options[] = {
    { "control-rate", "Rate Control Mode", OFFSET(control_rate), AV_OPT_TYPE_INT, { .i64 = 1}, 0,  5, VE, "control-rate"},
    { "max-bitrate", "Maximum Bit Rate", OFFSET(max_bitrate), AV_OPT_TYPE_INT64, { .i64 = 5000000}, 0,  35000000000, VE, "max-bitrate"},
    { "slice-qp", "Slice QP", OFFSET(slice_qp), AV_OPT_TYPE_INT, { .i64 = -1}, -1,  51, VE, "slice-qp"},
    { "min-qp", "Minimum QP value allowed for the rate control", OFFSET(min_qp), AV_OPT_TYPE_INT, { .i64 = 0}, 0, 51, VE, "min-qp"},
    { "max-qp", "Maximum QP value allowed for the rate control", OFFSET(max_qp), AV_OPT_TYPE_INT, { .i64 = 51}, 0, 51, VE, "max-qp"},
    { "cpb-size", "Coded Picture Buffer as specified in the HRD model in sec", OFFSET(cpb_size), AV_OPT_TYPE_DOUBLE, { .dbl = 2.0}, 0, 4294967295, VE, "cpb-size"},
    { "initial-delay", "Initial Removal Delay as specified in the HRD model in sec", OFFSET(initial_delay), AV_OPT_TYPE_DOUBLE, { .dbl = 1.0}, 0, 4294967295, VE, "initial-delay"},
    { "gop-mode", "GOP Configuration", OFFSET(gop_mode), AV_OPT_TYPE_INT, { .i64 = 0}, 0, 5, VE, "gop-mode"},
    { "gdr-mode", "Gradual Decoder Refresh scheme mode", OFFSET(gdr_mode), AV_OPT_TYPE_INT, { .i64 = 0}, 0, 2, VE, "gdr-mode"},
    { "bf", "Number of B-frames", OFFSET(b_frames), AV_OPT_TYPE_INT, { .i64 = 2}, 0, 4294967295, VE, "b-frames"},
    { "periodicity-idr", "IDR Picture Frequency", OFFSET(periodicity_idr), AV_OPT_TYPE_INT, { .i64 = 4294967295}, 0, 4294967295, VE, "periodicity-idr"},
    { "profile", "Set the encoding profile", OFFSET(profile), AV_OPT_TYPE_INT, { .i64 = FF_PROFILE_H264_HIGH }, FF_PROFILE_H264_BASELINE, FF_PROFILE_H264_HIGH_422_INTRA, VE, "profile" },
    { "level", "Set the encoding level restriction", OFFSET(level), AV_OPT_TYPE_INT, { .i64 = 50 }, 10, 52, VE, "level" },
    { "slices", "Number of Slices", OFFSET(num_slices), AV_OPT_TYPE_INT, { .i64 = 1}, 1, 68, VE, "slices"},
    { "qp-mode", "QP Control Mode", OFFSET(qp_mode), AV_OPT_TYPE_INT, { .i64 = 1}, 0, 3, VE, "qp-mode"},
    { "filler-data", "Specifies if filler data can be added to the stream or not", OFFSET(filler_data), AV_OPT_TYPE_INT, { .i64 = 1}, 0, 1, VE, "filler-data"},
    { "aspect-ratio", "Aspect-Ratio", OFFSET(aspect_ratio), AV_OPT_TYPE_INT, { .i64 = 0}, 0, 3, VE, "aspect-ratio"},
    { "slice-size", "If set to 0, slices are defined by the NumSlices parameter, Otherwise it specifies the target slice size, in bytes, that the encoder uses to automatically split the bitstream into approximately equally sized slices, with a granularity of one LCU.", OFFSET(slice_size), AV_OPT_TYPE_INT, { .i64 = 0}, 0, 65535, VE, "slice-size"},
    { "scaling-list", "Scaling List Mode", OFFSET(scaling_list), AV_OPT_TYPE_INT, { .i64 = 1}, 0, 1, VE, "scaling-list"},
    { "entropy-mode", "Entropy Mode", OFFSET(entropy_mode), AV_OPT_TYPE_INT, { .i64 = 1}, 0, 1, VE, "entropy-mode"},
    { "loop-filter", "Deblocking Filter", OFFSET(loop_filter), AV_OPT_TYPE_INT, { .i64 = 1}, 0, 1, VE, "loop-filter"},
    { "constrained-intra-pred", "Constrained Intra Prediction", OFFSET(constrained_intra_pred), AV_OPT_TYPE_INT, { .i64 = 0}, 0, 1, VE, "constrained-intra-pred"},
    { "prefetch-buffer", "Enable/Disable L2Cache buffer in encoding process", OFFSET(prefetch_buffer), AV_OPT_TYPE_INT, { .i64 = 1}, 0, 1, VE, "prefetch-buffer"},
    { "cores", "Number of cores to use", OFFSET(cores), AV_OPT_TYPE_INT, { .i64 = 0}, 0, 4, VE, "cores"},
    { "lambda_file", "lambda hex file", OFFSET(lambda_file), AV_OPT_TYPE_STRING, { }, CHAR_MIN, CHAR_MAX, VE, "lambda_file"},
    { "lookahead_depth", "Number of frames to lookahead for qp maps generation or custom rate control. Up to 20", OFFSET(lookahead_depth), AV_OPT_TYPE_INT, {.i64 = 0}, 0, 20, VE, "lookahead_depth"},
    { "temporal-aq", "Enable Temporal AQ.", OFFSET(temporal_aq), AV_OPT_TYPE_INT, {.i64 = 0}, 0, 1, VE, "temporal-aq-mode"},
    { "spatial-aq", "Enable Spatial AQ.", OFFSET(spatial_aq), AV_OPT_TYPE_INT, {.i64 = 0}, 0, 1, VE, "spatial-aq-mode"},
    { "rate-control-mode", "Custom rate control mode", OFFSET(rate_control_mode), AV_OPT_TYPE_INT, {.i64 = 0}, 0, 1, VE, "rate-control-mode"},
    { "spatial-aq-gain", "Percentage of spatial AQ gain", OFFSET(spatial_aq_gain), AV_OPT_TYPE_INT, {.i64 = 50}, 0, 100, VE, "spatial-aq-gain"},
	{ "latency_logging", "Log latency information to syslog", OFFSET(latency_logging), AV_OPT_TYPE_INT, { .i64 = 0 }, 0, 1, VE, "latency_logging" },

    { "const-qp", "Constant QP", 0, AV_OPT_TYPE_CONST, { .i64 = 0}, 0, 0, VE, "control-rate"},
    { "cbr", "Constant Bitrate", 0, AV_OPT_TYPE_CONST, { .i64 = 1}, 0, 0, VE, "control-rate"},
    { "vbr", "Variable Bitrate", 0, AV_OPT_TYPE_CONST, { .i64 = 2}, 0, 0, VE, "control-rate"},
    { "low-latency", "Low Latency", 0, AV_OPT_TYPE_CONST, { .i64 = 3}, 0, 0, VE, "control-rate"},
    { "capped-vbr", "Capped Variable Bitrate", 0, AV_OPT_TYPE_CONST, { .i64 = 4}, 0, 0, VE, "control-rate"},
    { "bypass", "Bypass", 0, AV_OPT_TYPE_CONST, { .i64 = 5}, 0, 0, VE, "control-rate"},
    { "auto", "Auto", 0, AV_OPT_TYPE_CONST, { .i64 = -1}, 0, 0, VE, "slice-qp"},
    { "default-gop", "Default GOP settings", 0, AV_OPT_TYPE_CONST, { .i64 = 0}, 0, 0, VE, "gop-mode"},
    { "pyramidal-gop", "Advanced GOP pattern with hierachical B-frames", 0, AV_OPT_TYPE_CONST, { .i64 = 1}, 0, 0, VE, "gop-mode"},
    { "adaptive-gop", "Advanced GOP pattern with adaptive B-frames", 0, AV_OPT_TYPE_CONST, { .i64 = 2}, 0, 0, VE, "gop-mode"},
    { "bypass", "Bypass", 0, AV_OPT_TYPE_CONST, { .i64 = 3}, 0, 0, VE, "gop-mode"},
    { "low-delay-p", "Single I-frame followed by P-frames only", 0, AV_OPT_TYPE_CONST, { .i64 = 4}, 0, 0, VE, "gop-mode"},
    { "low-delay-b", "Single I-frame followed by B-frames only", 0, AV_OPT_TYPE_CONST, { .i64 = 5}, 0, 0, VE, "gop-mode"},
    { "disabled", "No GDR", 0, AV_OPT_TYPE_CONST, { .i64 = 0}, 0, 0, VE, "gdr-mode"},
    { "vertical", "Gradual refresh using a vertical bar moving from left to right", 0, AV_OPT_TYPE_CONST, { .i64 = 1}, 0, 0, VE, "gdr-mode"},
    { "horizontal", "Gradual refresh using a horizontal bar moving from top to bottom", 0, AV_OPT_TYPE_CONST, { .i64 = 2}, 0, 0, VE, "gdr-mode"},
    { "baseline", "Baseline profile", 0, AV_OPT_TYPE_CONST, { .i64 = FF_PROFILE_H264_BASELINE}, 0, 0, VE, "profile"},
    { "main", "Main profile", 0, AV_OPT_TYPE_CONST, { .i64 = FF_PROFILE_H264_MAIN}, 0, 0, VE, "profile"},
    { "high", "High profile", 0, AV_OPT_TYPE_CONST, { .i64 = FF_PROFILE_H264_HIGH}, 0, 0, VE, "profile"},
    { "high-10", "High 10 profile", 0, AV_OPT_TYPE_CONST, { .i64 = FF_PROFILE_H264_HIGH_10}, 0, 0, VE, "profile"},
    { "high-422", "High 4:2:2 profile", 0, AV_OPT_TYPE_CONST, { .i64 = FF_PROFILE_H264_HIGH_422}, 0, 0, VE, "profile"},
    { "high-10-intra", "High 10 Intra profile", 0, AV_OPT_TYPE_CONST, { .i64 = FF_PROFILE_H264_HIGH_10_INTRA}, 0, 0, VE, "profile"},
    { "high-422-intra", "High 4:2:2 Intra profile", 0, AV_OPT_TYPE_CONST, { .i64 = FF_PROFILE_H264_HIGH_422_INTRA}, 0, 0, VE, "profile"},
    { "1", "1 level", 0, AV_OPT_TYPE_CONST, { .i64 = 10}, 0, 0, VE, "level"},
    { "1.1", "1.1 level", 0, AV_OPT_TYPE_CONST, { .i64 = 11}, 0, 0, VE, "level"},
    { "1.2", "1.2 level", 0, AV_OPT_TYPE_CONST, { .i64 = 12}, 0, 0, VE, "level"},
    { "1.3", "1.3 level", 0, AV_OPT_TYPE_CONST, { .i64 = 13}, 0, 0, VE, "level"},
    { "2", "2 level", 0, AV_OPT_TYPE_CONST, { .i64 = 20}, 0, 0, VE, "level"},
    { "2.1", "2.1 level", 0, AV_OPT_TYPE_CONST, { .i64 = 21}, 0, 0, VE, "level"},
    { "2.2", "2.2 level", 0, AV_OPT_TYPE_CONST, { .i64 = 22}, 0, 0, VE, "level"},
    { "3", "3 level", 0, AV_OPT_TYPE_CONST, { .i64 = 30}, 0, 0, VE, "level"},
    { "3.1", "3.1 level", 0, AV_OPT_TYPE_CONST, { .i64 = 31}, 0, 0, VE, "level"},
    { "3.2", "3.2 level", 0, AV_OPT_TYPE_CONST, { .i64 = 32}, 0, 0, VE, "level"},
    { "4", "4 level", 0, AV_OPT_TYPE_CONST, { .i64 = 40}, 0, 0, VE, "level"},
    { "4.1", "4.1 level", 0, AV_OPT_TYPE_CONST, { .i64 = 41}, 0, 0, VE, "level"},
    { "4.2", "4.2 level", 0, AV_OPT_TYPE_CONST, { .i64 = 42}, 0, 0, VE, "level"},
    { "5", "5 level", 0, AV_OPT_TYPE_CONST, { .i64 = 50}, 0, 0, VE, "level"},
    { "5.1", "5.1 level", 0, AV_OPT_TYPE_CONST, { .i64 = 51}, 0, 0, VE, "level"},
    { "5.2", "5.2 level", 0, AV_OPT_TYPE_CONST, { .i64 = 52}, 0, 0, VE, "level"},
    { "uniform", "Use the same QP for all coding units of the frame", 0, AV_OPT_TYPE_CONST, { .i64 = 0}, 0, 0, VE, "qp-mode"},
    { "auto", "Let the VCU encoder change the QP for each coding unit according to its content", 0, AV_OPT_TYPE_CONST, { .i64 = 1}, 0, 0, VE, "qp-mode"},
    { "load", "Use the values in the lambda file", 0, AV_OPT_TYPE_CONST, { .i64 = 2}, 0, 0, VE, "qp-mode"},
    { "relative-load", "Use the information gathered in the lookahead to calculate the best QP", 0, AV_OPT_TYPE_CONST, { .i64 = 3}, 0, 0, VE, "qp-mode"},
    { "enable", "Enable filler data", 0, AV_OPT_TYPE_CONST, { .i64 = 1}, 0, 0, VE, "filler-data"},
    { "disable", "Disable filler data", 0, AV_OPT_TYPE_CONST, { .i64 = 0}, 0, 0, VE, "filler-data"},
    { "auto", "4:3 for SD video, 16:9 for HD video, unspecified for unknown format", 0, AV_OPT_TYPE_CONST, { .i64 = 0}, 0, 0, VE, "aspect-ratio"},
    { "4:3", "4:3 aspect ratio", 0, AV_OPT_TYPE_CONST, { .i64 = 1}, 0, 0, VE, "aspect-ratio"},
    { "16:9", "16:9 aspect ratio", 0, AV_OPT_TYPE_CONST, { .i64 = 2}, 0, 0, VE, "aspect-ratio"},
    { "none", "Aspect ratio information is not present in the stream", 0, AV_OPT_TYPE_CONST, { .i64 = 3}, 0, 0, VE, "aspect-ratio"},
    { "flat", "Flat scaling list mode", 0, AV_OPT_TYPE_CONST, { .i64 = 0}, 0, 0, VE, "scaling-list"},
    { "default", "Default scaling list mode", 0, AV_OPT_TYPE_CONST, { .i64 = 1}, 0, 0, VE, "scaling-list"},
    { "CAVLC", "CAVLC entropy mode", 0, AV_OPT_TYPE_CONST, { .i64 = 0}, 0, 0, VE, "entropy-mode"},
    { "CABAC", "CABAC entropy mode", 0, AV_OPT_TYPE_CONST, { .i64 = 1}, 0, 0, VE, "entropy-mode"},
    { "enable", "Enable deblocking filter", 0, AV_OPT_TYPE_CONST, { .i64 = 1}, 0, 0, VE, "loop-filter"},
    { "disable", "Disable deblocking filter", 0, AV_OPT_TYPE_CONST, { .i64 = 0}, 0, 0, VE, "loop-filter"},
    { "enable", "Enable constrained intra prediction", 0, AV_OPT_TYPE_CONST, { .i64 = 1}, 0, 0, VE, "constrained-intra-pred"},
    { "disable", "Disable constrained intra prediction", 0, AV_OPT_TYPE_CONST, { .i64 = 0}, 0, 0, VE, "constrained-intra-pred"},
    { "enable", "Enable L2 caching", 0, AV_OPT_TYPE_CONST, { .i64 = 1}, 0, 0, VE, "prefetch-buffer"},
    { "disable", "Disable L2 caching", 0, AV_OPT_TYPE_CONST, { .i64 = 0}, 0, 0, VE, "prefetch-buffer"},
    { "auto", "Automatic", 0, AV_OPT_TYPE_CONST, { .i64 = 0}, 0, 0, VE, "cores"},
    { "disable", "Disable Temporal AQ", 0, AV_OPT_TYPE_CONST, { .i64 = 0}, 0, 0, VE, "temporal-aq-mode"},
    { "enable", "Enable Temporal AQ", 0, AV_OPT_TYPE_CONST, { .i64 = 1}, 0, 0, VE, "temporal-aq-mode"},
    { "disable", "Disable Spatial AQ", 0, AV_OPT_TYPE_CONST, { .i64 = 0}, 0, 0, VE, "spatial-aq-mode"},
    { "enable", "Enable Spatial AQ", 0, AV_OPT_TYPE_CONST, { .i64 = 1}, 0, 0, VE, "spatial-aq-mode"},
    { "disable", "Disable Custom Rate Control", 0, AV_OPT_TYPE_CONST, { .i64 = 0}, 0, 0, VE, "rate-control-mode"},
    { "enable", "Enable Custom Rate Control", 0, AV_OPT_TYPE_CONST, { .i64 = 1}, 0, 0, VE, "rate-control-mode"},
    {NULL},
};

static const AVOption hevcOptions[] = {
    { "control-rate", "Rate Control Mode", OFFSET(control_rate), AV_OPT_TYPE_INT, { .i64 = 1}, 0,  5, VE, "control-rate"},
    { "max-bitrate", "Maximum Bit Rate", OFFSET(max_bitrate), AV_OPT_TYPE_INT64, { .i64 = 5000000}, 0,  35000000000, VE, "max-bitrate"},
    { "slice-qp", "Slice QP", OFFSET(slice_qp), AV_OPT_TYPE_INT, { .i64 = -1}, -1,  51, VE, "slice-qp"},
    { "min-qp", "Minimum QP value allowed for the rate control", OFFSET(min_qp), AV_OPT_TYPE_INT, { .i64 = 0}, 0, 51, VE, "min-qp"},
    { "max-qp", "Maximum QP value allowed for the rate control", OFFSET(max_qp), AV_OPT_TYPE_INT, { .i64 = 51}, 0, 51, VE, "max-qp"},
    { "cpb-size", "Coded Picture Buffer as specified in the HRD model in sec", OFFSET(cpb_size), AV_OPT_TYPE_DOUBLE, { .dbl = 2.0}, 0, 4294967295, VE, "cpb-size"},
    { "initial-delay", "Initial Removal Delay as specified in the HRD model in sec", OFFSET(initial_delay), AV_OPT_TYPE_DOUBLE, { .dbl = 1.0}, 0, 4294967295, VE, "initial-delay"},
    { "gop-mode", "GOP Configuration", OFFSET(gop_mode), AV_OPT_TYPE_INT, { .i64 = 0}, 0, 5, VE, "gop-mode"},
    { "gdr-mode", "Gradual Decoder Refresh scheme mode", OFFSET(gdr_mode), AV_OPT_TYPE_INT, { .i64 = 0}, 0, 2, VE, "gdr-mode"},
    { "bf", "Number of B-frames", OFFSET(b_frames), AV_OPT_TYPE_INT, { .i64 = 2}, 0, 4294967295, VE, "b-frames"},
    { "periodicity-idr", "IDR Picture Frequency", OFFSET(periodicity_idr), AV_OPT_TYPE_INT, { .i64 = 4294967295}, 0, 4294967295, VE, "periodicity-idr"},
    { "profile", "Set the encoding profile", OFFSET(profile), AV_OPT_TYPE_INT, { .i64 = 0 }, 0, 6, VE, "profile" },
    { "level", "Set the encoding level restriction", OFFSET(level), AV_OPT_TYPE_INT, { .i64 = 50 }, 10, 51, VE, "level" },
    { "tier", "Set the encoding tier", OFFSET(tier), AV_OPT_TYPE_INT, { .i64 = 0}, 0, 1, VE, "tier" },
    { "slices", "Number of Slices", OFFSET(num_slices), AV_OPT_TYPE_INT, { .i64 = 1}, 1, 68, VE, "slices"},
    { "qp-mode", "QP Control Mode", OFFSET(qp_mode), AV_OPT_TYPE_INT, { .i64 = 1}, 0, 3, VE, "qp-mode"},
    { "filler-data", "Specifies if filler data can be added to the stream or not", OFFSET(filler_data), AV_OPT_TYPE_INT, { .i64 = 1}, 0, 1, VE, "filler-data"},
    { "aspect-ratio", "Aspect-Ratio", OFFSET(aspect_ratio), AV_OPT_TYPE_INT, { .i64 = 0}, 0, 3, VE, "aspect-ratio"},
    { "dependent-slice", "When there are several slices per frames, this parameter specifies whether the additional slices are dependent slice segments or regular slices", OFFSET(dependent_slice), AV_OPT_TYPE_INT, { .i64 = 0}, 0,  1, VE, "dependent-slice"},
    { "slice-size", "If set to 0, slices are defined by the NumSlices parameter, Otherwise it specifies the target slice size, in bytes, that the encoder uses to automatically split the bitstream into approximately equally sized slices, with a granularity of one LCU.", OFFSET(slice_size), AV_OPT_TYPE_INT, { .i64 = 0}, 0, 65535, VE, "slice-size"},
    { "scaling-list", "Scaling List Mode", OFFSET(scaling_list), AV_OPT_TYPE_INT, { .i64 = 1}, 0, 1, VE, "scaling-list"},
    { "loop-filter", "Deblocking Filter", OFFSET(loop_filter), AV_OPT_TYPE_INT, { .i64 = 1}, 0, 1, VE, "loop-filter"},
    { "constrained-intra-pred", "Constrained Intra Prediction", OFFSET(constrained_intra_pred), AV_OPT_TYPE_INT, { .i64 = 0}, 0, 1, VE, "constrained-intra-pred"},
    { "prefetch-buffer", "Enable/Disable L2Cache buffer in encoding process", OFFSET(prefetch_buffer), AV_OPT_TYPE_INT, { .i64 = 1}, 0, 1, VE, "prefetch-buffer"},
    { "cores", "Number of cores to use", OFFSET(cores), AV_OPT_TYPE_INT, { .i64 = 0}, 0, 4, VE, "cores"},
    { "lambda_file", "lambda hex file", OFFSET(lambda_file), AV_OPT_TYPE_STRING, { }, CHAR_MIN, CHAR_MAX, VE, "lambda_file"},
    { "lookahead_depth", "Number of frames to lookahead for qp maps generation or custom rate control. Up to 20", OFFSET(lookahead_depth), AV_OPT_TYPE_INT, {.i64 = 0}, 0, 20, VE, "lookahead_depth"},
    { "temporal-aq", "Enable Temporal AQ.", OFFSET(temporal_aq), AV_OPT_TYPE_INT, {.i64 = 0}, 0, 1, VE, "temporal-aq-mode"},
    { "spatial-aq", "Enable Spatial AQ.", OFFSET(spatial_aq), AV_OPT_TYPE_INT, {.i64 = 0}, 0, 1, VE, "spatial-aq-mode"},
    { "rate-control-mode", "Custom rate control mode", OFFSET(rate_control_mode), AV_OPT_TYPE_INT, {.i64 = 0}, 0, 1, VE, "rate-control-mode"},
    { "spatial-aq-gain", "Percentage of spatial AQ gain", OFFSET(spatial_aq_gain), AV_OPT_TYPE_INT, {.i64 = 50}, 0, 100, VE, "spatial-aq-gain"},
	{ "latency_logging", "Log latency information to syslog", OFFSET(latency_logging), AV_OPT_TYPE_INT, { .i64 = 0 }, 0, 1, VE, "latency_logging" },

    { "const-qp", "Constant QP", 0, AV_OPT_TYPE_CONST, { .i64 = 0}, 0, 0, VE, "control-rate"},
    { "cbr", "Constant Bitrate", 0, AV_OPT_TYPE_CONST, { .i64 = 1}, 0, 0, VE, "control-rate"},
    { "vbr", "Variable Bitrate", 0, AV_OPT_TYPE_CONST, { .i64 = 2}, 0, 0, VE, "control-rate"},
    { "low-latency", "Low Latency", 0, AV_OPT_TYPE_CONST, { .i64 = 3}, 0, 0, VE, "control-rate"},
    { "capped-vbr", "Capped Variable Bitrate", 0, AV_OPT_TYPE_CONST, { .i64 = 4}, 0, 0, VE, "control-rate"},
    { "bypass", "Bypass", 0, AV_OPT_TYPE_CONST, { .i64 = 5}, 0, 0, VE, "control-rate"},
    { "auto", "Auto", 0, AV_OPT_TYPE_CONST, { .i64 = -1}, 0, 0, VE, "slice-qp"},
    { "default-gop", "Default GOP settings", 0, AV_OPT_TYPE_CONST, { .i64 = 0}, 0, 0, VE, "gop-mode"},
    { "pyramidal-gop", "Advanced GOP pattern with hierachical B-frames", 0, AV_OPT_TYPE_CONST, { .i64 = 1}, 0, 0, VE, "gop-mode"},
    { "adaptive-gop", "Advanced GOP pattern with adaptive B-frames", 0, AV_OPT_TYPE_CONST, { .i64 = 2}, 0, 0, VE, "gop-mode"},
    { "bypass", "Bypass", 0, AV_OPT_TYPE_CONST, { .i64 = 3}, 0, 0, VE, "gop-mode"},
    { "low-delay-p", "Single I-frame followed by P-frames only", 0, AV_OPT_TYPE_CONST, { .i64 = 4}, 0, 0, VE, "gop-mode"},
    { "low-delay-b", "Single I-frame followed by B-frames only", 0, AV_OPT_TYPE_CONST, { .i64 = 5}, 0, 0, VE, "gop-mode"},
    { "disabled", "No GDR", 0, AV_OPT_TYPE_CONST, { .i64 = 0}, 0, 0, VE, "gdr-mode"},
    { "vertical", "Gradual refresh using a vertical bar moving from left to right", 0, AV_OPT_TYPE_CONST, { .i64 = 1}, 0, 0, VE, "gdr-mode"},
    { "horizontal", "Gradual refresh using a horizontal bar moving from top to bottom", 0, AV_OPT_TYPE_CONST, { .i64 = 2}, 0, 0, VE, "gdr-mode"},
    { "main", "Main profile", 0, AV_OPT_TYPE_CONST, { .i64 = 0}, 0, 0, VE, "profile"},
    { "main-intra", "Main Intra profile", 0, AV_OPT_TYPE_CONST, { .i64 = 1}, 0, 0, VE, "profile"},
    { "main-10", "Main 10 profile", 0, AV_OPT_TYPE_CONST, { .i64 = 2}, 0, 0, VE, "profile"},
    { "main-10-intra", "Main 10 Intra profile", 0, AV_OPT_TYPE_CONST, { .i64 = 3}, 0, 0, VE, "profile"},
    { "main-422", "Main 4:2:2 profile", 0, AV_OPT_TYPE_CONST, { .i64 = 4}, 0, 0, VE, "profile"},
    { "main-422-10", "Main 4:2:2 10 profile", 0, AV_OPT_TYPE_CONST, { .i64 = 5}, 0, 0, VE, "profile"},
    { "main-422-10-intra", "Main 4:2:2 10 Intra profile", 0, AV_OPT_TYPE_CONST, { .i64 = 6}, 0, 0, VE, "profile"},
    { "1", "1 level", 0, AV_OPT_TYPE_CONST, { .i64 = 10}, 0, 0, VE, "level"},
    { "2", "2 level", 0, AV_OPT_TYPE_CONST, { .i64 = 20}, 0, 0, VE, "level"},
    { "2.1", "2.1 level", 0, AV_OPT_TYPE_CONST, { .i64 = 21}, 0, 0, VE, "level"},
    { "3", "3 level", 0, AV_OPT_TYPE_CONST, { .i64 = 30}, 0, 0, VE, "level"},
    { "3.1", "3.1 level", 0, AV_OPT_TYPE_CONST, { .i64 = 31}, 0, 0, VE, "level"},
    { "4", "4 level", 0, AV_OPT_TYPE_CONST, { .i64 = 40}, 0, 0, VE, "level"},
    { "4.1", "4.1 level", 0, AV_OPT_TYPE_CONST, { .i64 = 41}, 0, 0, VE, "level"},
    { "5", "5 level", 0, AV_OPT_TYPE_CONST, { .i64 = 50}, 0, 0, VE, "level"},
    { "5.1", "5.1 level", 0, AV_OPT_TYPE_CONST, { .i64 = 51}, 0, 0, VE, "level"},
    { "main", "Main tier", 0, AV_OPT_TYPE_CONST, { .i64 = 0}, 0, 0, VE, "tier"},
    { "high", "High tier", 0, AV_OPT_TYPE_CONST, { .i64 = 1}, 0, 0, VE, "tier"},
    { "uniform", "Use the same QP for all coding units of the frame", 0, AV_OPT_TYPE_CONST, { .i64 = 0}, 0, 0, VE, "qp-mode"},
    { "auto", "Let the VCU encoder change the QP for each coding unit according to its content", 0, AV_OPT_TYPE_CONST, { .i64 = 1}, 0, 0, VE, "qp-mode"},
    { "load", "Use the values in the lambda file", 0, AV_OPT_TYPE_CONST, { .i64 = 2}, 0, 0, VE, "qp-mode"},
    { "relative-load", "Use the information gathered in the lookahead to calculate the best QP", 0, AV_OPT_TYPE_CONST, { .i64 = 3}, 0, 0, VE, "qp-mode"},
    { "enable", "Enable filler data", 0, AV_OPT_TYPE_CONST, { .i64 = 1}, 0, 0, VE, "filler-data"},
    { "disable", "Disable filler data", 0, AV_OPT_TYPE_CONST, { .i64 = 0}, 0, 0, VE, "filler-data"},
    { "auto", "4:3 for SD video, 16:9 for HD video, unspecified for unknown format", 0, AV_OPT_TYPE_CONST, { .i64 = 0}, 0, 0, VE, "aspect-ratio"},
    { "4:3", "4:3 aspect ratio", 0, AV_OPT_TYPE_CONST, { .i64 = 1}, 0, 0, VE, "aspect-ratio"},
    { "16:9", "16:9 aspect ratio", 0, AV_OPT_TYPE_CONST, { .i64 = 2}, 0, 0, VE, "aspect-ratio"},
    { "none", "Aspect ratio information is not present in the stream", 0, AV_OPT_TYPE_CONST, { .i64 = 3}, 0, 0, VE, "aspect-ratio"},
    { "enable", "Enable dependent slices", 0, AV_OPT_TYPE_CONST, { .i64 = 1}, 0, 0, VE, "dependent-slice"},
    { "disable", "Disable dependent slices", 0, AV_OPT_TYPE_CONST, { .i64 = 0}, 0, 0, VE, "dependent-slice"},
    { "flat", "Flat scaling list mode", 0, AV_OPT_TYPE_CONST, { .i64 = 0}, 0, 0, VE, "scaling-list"},
    { "default", "Default scaling list mode", 0, AV_OPT_TYPE_CONST, { .i64 = 1}, 0, 0, VE, "scaling-list"},
    { "enable", "Enable deblocking filter", 0, AV_OPT_TYPE_CONST, { .i64 = 1}, 0, 0, VE, "loop-filter"},
    { "disable", "Disable deblocking filter", 0, AV_OPT_TYPE_CONST, { .i64 = 0}, 0, 0, VE, "loop-filter"},
    { "enable", "Enable constrained intra prediction", 0, AV_OPT_TYPE_CONST, { .i64 = 1}, 0, 0, VE, "constrained-intra-pred"},
    { "disable", "Disable constrained intra prediction", 0, AV_OPT_TYPE_CONST, { .i64 = 0}, 0, 0, VE, "constrained-intra-pred"},
    { "enable", "Enable L2 caching", 0, AV_OPT_TYPE_CONST, { .i64 = 1}, 0, 0, VE, "prefetch-buffer"},
    { "disable", "Disable L2 caching", 0, AV_OPT_TYPE_CONST, { .i64 = 0}, 0, 0, VE, "prefetch-buffer"},
    { "auto", "Automatic", 0, AV_OPT_TYPE_CONST, { .i64 = 0}, 0, 0, VE, "cores"},
    { "disable", "Disable Temporal AQ", 0, AV_OPT_TYPE_CONST, { .i64 = 0}, 0, 0, VE, "temporal-aq-mode"},
    { "enable", "Enable Temporal AQ", 0, AV_OPT_TYPE_CONST, { .i64 = 1}, 0, 0, VE, "temporal-aq-mode"},
    { "disable", "Disable Spatial AQ", 0, AV_OPT_TYPE_CONST, { .i64 = 0}, 0, 0, VE, "spatial-aq-mode"},
    { "enable", "Enable Spatial AQ", 0, AV_OPT_TYPE_CONST, { .i64 = 1}, 0, 0, VE, "spatial-aq-mode"},
    { "disable", "Disable Custom Rate Control", 0, AV_OPT_TYPE_CONST, { .i64 = 0}, 0, 0, VE, "rate-control-mode"},
    { "enable", "Enable Custom Rate Control", 0, AV_OPT_TYPE_CONST, { .i64 = 1}, 0, 0, VE, "rate-control-mode"},
    {NULL},
};

static void
mpsoc_vcu_encode_queue_pts (AVFifoBuffer* queue, int64_t pts)
{
    av_fifo_generic_write(queue, &pts, sizeof(pts), NULL);
}

static int64_t
mpsoc_vcu_encode_dequeue_pts(AVFifoBuffer* queue)
{
    int64_t ts = AV_NOPTS_VALUE;
    if (av_fifo_size(queue) > 0)
        av_fifo_generic_read(queue, &ts, sizeof(ts), NULL);
    return ts;
}

static bool mpsoc_encode_is_h264_idr(AVPacket *pkt)
{
    unsigned char* pt = pkt->data;
    unsigned char* end = pkt->data + pkt->size - 3;
    while (pt < end)
    {
        if ((pt[0] == 0x00) && (pt[1] == 0x00) && (pt[2] == 0x01) && ((pt[3] & 0x1F) == 0x05))
            return true;
        pt++;
    }
    return false;
}

static bool mpsoc_encode_is_hevc_idr(AVPacket *pkt)
{
    unsigned char* pt = pkt->data;
    unsigned char* end = pkt->data + pkt->size - 3;
    while (pt < end)
    {
        if ((pt[0] == 0x00) && (pt[1] == 0x00) && (pt[2] == 0x01))
        {
            unsigned char naluType = (pt[3] & 0x7E) >> 1;
            if (naluType == 19 || naluType == 20)
                return true;
        }
        pt++;
    }
    return false;
}

static void
mpsoc_vcu_encode_prepare_out_timestamp (AVCodecContext *avctx, AVPacket *pkt)
{
    mpsoc_vcu_enc_ctx *ctx = avctx->priv_data;

    // TODO: code is written based on nvenc, need to check legalities of this
    if (ctx->pts_1 != AV_NOPTS_VALUE && ctx->b_frames > 0 && ctx->is_first_outframe) {
        int64_t ts0 = ctx->pts_0, ts1 = ctx->pts_1;
        int64_t delta;

        if ((ts0 < 0 && ts1 > INT64_MAX + ts0) ||
            (ts0 > 0 && ts1 < INT64_MIN + ts0))
            return;
        delta = ts1 - ts0;

        if ((delta < 0 && ts0 > INT64_MAX + delta) ||
            (delta > 0 && ts0 < INT64_MIN + delta))
          return;
        pkt->dts = ts0 - delta;

        ctx->is_first_outframe = 0;
        return;
    }

    pkt->dts = mpsoc_vcu_encode_dequeue_pts(ctx->pts_queue);
}

static void deinit_la(mpsoc_vcu_enc_ctx *ctx)
{
    XmaFrame *xmaframe;
    if (!ctx->la) {
        return;
    }

    destroy_xlnx_la(ctx->la);
    ctx->la = NULL;
    if (ctx->laQ) {
        while (av_fifo_size(ctx->laQ) >= sizeof(XmaFrame *)) {
            av_fifo_generic_read(ctx->laQ, &xmaframe, sizeof(XmaFrame *), NULL);
            if (xmaframe->data[0].buffer_type == XMA_HOST_BUFFER_TYPE) {
            	xmaframe->data[0].buffer = NULL;
            	xmaframe->data[1].buffer = NULL;
            }
            xma_frame_clear_all_side_data(xmaframe);
            av_free(xmaframe);
            xmaframe = NULL;
        }
        av_fifo_free(ctx->laQ);
        ctx->laQ = NULL;
    }
}

static int32_t init_la(AVCodecContext *avctx)
{
    xlnx_la_cfg_t la_cfg;
    mpsoc_vcu_enc_ctx *ctx = avctx->priv_data;
    la_cfg.width = avctx->width;
    la_cfg.height = avctx->height;
    la_cfg.framerate.numerator = avctx->framerate.num;
    la_cfg.framerate.denominator = avctx->framerate.den;
    //@TODO: Assume 256 aligned for now. Needs to be fixed later
    la_cfg.stride = FFALIGN(avctx->width, 256);
    la_cfg.bits_per_pixel = 8;

    if (avctx->gop_size <= 0) {
        la_cfg.gop_size = 120;
    } else {
        la_cfg.gop_size = avctx->gop_size;
    }

    la_cfg.lookahead_depth = ctx->lookahead_depth;
    la_cfg.spatial_aq_mode = ctx->spatial_aq;
    la_cfg.spatial_aq_gain = ctx->spatial_aq_gain;
    la_cfg.temporal_aq_mode = ctx->temporal_aq;
    la_cfg.rate_control_mode = ctx->rate_control_mode;
    la_cfg.b_frames = ctx->b_frames;
    la_cfg.framerate.numerator   = avctx->framerate.num;
    la_cfg.framerate.denominator = avctx->framerate.den;
    la_cfg.latency_logging = ctx->latency_logging;
    switch (avctx->pix_fmt) {
    case AV_PIX_FMT_NV12:
        la_cfg.enableZeroCopy = 0;
        la_cfg.fmt_type = XMA_VCU_NV12_FMT_TYPE;
        break;
    case AV_PIX_FMT_VCU_NV12:
        la_cfg.enableZeroCopy = 1;
        la_cfg.fmt_type = XMA_VCU_NV12_FMT_TYPE;
        break;
    case AV_PIX_FMT_YUV420P:
        la_cfg.enableZeroCopy = 0;
        la_cfg.fmt_type = XMA_YUV420_FMT_TYPE;
        break;
    }
    switch (avctx->codec_id) {
    case AV_CODEC_ID_H264:
        la_cfg.codec_type = EXlnxAvc;
        break;
    case AV_CODEC_ID_HEVC:
        la_cfg.codec_type = EXlnxHevc;
        break;
    }

    ctx->la = create_xlnx_la(&la_cfg);
    if (!ctx->la) {
        printf("Error : init_la : create_xlnx_la Failed OOM\n");
        return AVERROR(ENOMEM);
    }
	ctx->laQ = av_fifo_alloc((ctx->lookahead_depth + 1) * sizeof(XmaFrame *));
	if (ctx->laQ == NULL) {
        printf("Error : init_la : av_fifo_alloc Failed OOM\n");
        deinit_la(ctx);
        return AVERROR(ENOMEM);
    }
    for (int i = 0; i < (ctx->lookahead_depth + 1); i++) {
        //@TODO for AV_PIX_FMT_VCU_NV12 alloc on XmaFrame size
        //XmaFrame *in_xma_frame = xma_frame_alloc(&ctx->frame.frame_props);
        XmaFrame *in_xma_frame = av_mallocz(sizeof(XmaFrame));
        if (in_xma_frame == NULL) {
            deinit_la(ctx);
            printf("Error : init_la : xma_frame_alloc Failed OOM\n");
            return AVERROR(ENOMEM);
        }
        memcpy (in_xma_frame, &ctx->frame, sizeof(XmaFrame));
        av_fifo_generic_write(ctx->laQ, &in_xma_frame, sizeof(XmaFrame *), NULL);
    }
    return 0;
}

static av_cold int mpsoc_vcu_encode_close(AVCodecContext *avctx)
{
    mpsoc_vcu_enc_ctx *ctx = avctx->priv_data;

    if (getenv("XRM_RESERVE_ID"))
    {
       int xrm_reserve_id = atoi(getenv("XRM_RESERVE_ID"));

       //XRM encoder de-allocation
       if (ctx->encode_res_inuse ==1)
       {
          if (!(xrmCuListRelease(ctx->xrm_ctx, &ctx->encode_cu_list_res)))
             printf("fail to release encode cu\n");

       }
    }
    av_fifo_freep(&ctx->pts_queue);
    if (xrmDestroyContext(ctx->xrm_ctx) != XRM_SUCCESS)
       printf("XRM : destroy context failed\n");

     xma_enc_session_destroy(ctx->enc_session);
    deinit_la(ctx);
    return 0;
}

static void fill_options_file_h264 (AVCodecContext *avctx)
{
    mpsoc_vcu_enc_ctx *ctx = avctx->priv_data;

    const char* RateCtrlMode = "CONST_QP";
    switch (ctx->control_rate) {
        case 0: RateCtrlMode = "CONST_QP"; break;
        case 1: RateCtrlMode = "CBR"; break;
        case 2: RateCtrlMode = "VBR"; break;
        case 3: RateCtrlMode = "LOW_LATENCY"; break;
        case 4: RateCtrlMode = "CAPPED_VBR"; break;
        case 5: RateCtrlMode = "BYPASS"; break;
    }

    char FrameRate[16];
    sprintf(FrameRate, "%llu/%llu", avctx->framerate.num, avctx->framerate.den);

    char SliceQP[8];
    if (ctx->slice_qp == -1)
        strcpy (SliceQP, "AUTO");
    else
        sprintf(SliceQP, "%d", ctx->slice_qp);

    const char* GopCtrlMode = "DEFAULT_GOP";
    switch (ctx->gop_mode) {
        case 0: GopCtrlMode = "DEFAULT_GOP"; break;
        case 1: GopCtrlMode = "PYRAMIDAL_GOP"; break;
        case 2: GopCtrlMode = "ADAPTIVE_GOP"; break;
        case 3: GopCtrlMode = "BYPASS"; break;
        case 4: GopCtrlMode = "LOW_DELAY_P"; break;
        case 5: GopCtrlMode = "LOW_DELAY_B"; break;
    }

    const char* GDRMode = "DISABLE";
    switch (ctx->gdr_mode) {
        case 0: GDRMode = "DISABLE"; break;
        case 1: GDRMode = "GDR_VERTICAL"; break;
        case 2: GDRMode = "GDR_HORIZONTAL"; break;
    }

    const char* Profile = "AVC_BASELINE";
    switch (ctx->profile) {
        case FF_PROFILE_H264_BASELINE: Profile = "AVC_BASELINE"; break;
        case FF_PROFILE_H264_MAIN: Profile = "AVC_MAIN"; break;
        case FF_PROFILE_H264_HIGH: Profile = "AVC_HIGH"; break;
        case FF_PROFILE_H264_HIGH_10: Profile = "AVC_HIGH10"; break;
        case FF_PROFILE_H264_HIGH_10_INTRA: Profile = "AVC_HIGH10_INTRA"; break;
        case FF_PROFILE_H264_HIGH_422: Profile = "AVC_HIGH_422"; break;
        case FF_PROFILE_H264_HIGH_422_INTRA: Profile = "AVC_HIGH_422_INTRA"; break;
    }

    const char* Level = "1";
    switch (ctx->level) {
        case 10: Level = "1"; break;
        case 11: Level = "1.1"; break;
        case 12: Level = "1.2"; break;
        case 13: Level = "1.3"; break;
        case 20: Level = "2"; break;
        case 21: Level = "2.1"; break;
        case 22: Level = "2.2"; break;
        case 30: Level = "3"; break;
        case 31: Level = "3.1"; break;
        case 32: Level = "3.2"; break;
        case 40: Level = "4"; break;
        case 41: Level = "4.1"; break;
        case 42: Level = "4.2"; break;
        case 50: Level = "5"; break;
        case 51: Level = "5.1"; break;
        case 52: Level = "5.2"; break;
    }

    const char* QPCtrlMode = "UNIFORM_QP";
    switch (ctx->qp_mode) {
        case 0: QPCtrlMode = "UNIFORM_QP"; break;
        case 1: QPCtrlMode = "AUTO_QP"; break;
        case 2: QPCtrlMode = "LOAD_QP"; break;
        case 3: QPCtrlMode = "LOAD_QP | RELATIVE_QP"; break;
    }

    const char* FillerData = "ENABLE";
    switch (ctx->filler_data) {
        case 0: FillerData = "DISABLE"; break;
        case 1: FillerData = "ENABLE"; break;
    }

    const char* AspectRatio = "ASPECT_RATIO_AUTO";
    switch (ctx->aspect_ratio) {
        case 0: AspectRatio = "ASPECT_RATIO_AUTO"; break;
        case 1: AspectRatio = "ASPECT_RATIO_4_3"; break;
        case 2: AspectRatio = "ASPECT_RATIO_16_9"; break;
        case 3: AspectRatio = "ASPECT_RATIO_NONE"; break;
    }

    const char* ColorSpace = "COLOUR_DESC_UNSPECIFIED";
    switch (avctx->colorspace) {
        case AVCOL_SPC_BT709: ColorSpace = "COLOUR_DESC_BT_709"; break;
        case AVCOL_SPC_UNSPECIFIED: ColorSpace = "COLOUR_DESC_UNSPECIFIED"; break;
        case AVCOL_SPC_RESERVED: ColorSpace = "COLOUR_DESC_RESERVED"; break;
        case AVCOL_SPC_BT470BG: ColorSpace = "COLOUR_DESC_BT_470_NTSC"; break;
        case AVCOL_SPC_SMPTE170M: ColorSpace = "COLOUR_DESC_BT_601_PAL"; break;
        case AVCOL_SPC_SMPTE240M: ColorSpace = "COLOUR_DESC_BT_601_NTSC"; break;
        case AVCOL_SPC_BT2020_NCL: ColorSpace = "COLOUR_DESC_BT_2020"; break;
        case AVCOL_SPC_BT2020_CL: ColorSpace = "COLOUR_DESC_BT_2020"; break;
    }

    const char* ScalingList = "FLAT";
    switch (ctx->scaling_list) {
        case 0: ScalingList = "FLAT"; break;
        case 1: ScalingList = "DEFAULT"; break;
    }

    const char* EntropyMode = "MODE_CABAC";
    switch (ctx->entropy_mode) {
        case 0: EntropyMode = "MODE_CAVLC"; break;
        case 1: EntropyMode = "MODE_CABAC"; break;
    }

    const char* LoopFilter = "ENABLE";
    switch (ctx->loop_filter) {
        case 0: LoopFilter = "DISABLE"; break;
        case 1: LoopFilter = "ENABLE"; break;
    }

    const char* ConstIntraPred = "ENABLE";
    switch (ctx->constrained_intra_pred) {
        case 0: ConstIntraPred = "DISABLE"; break;
        case 1: ConstIntraPred = "ENABLE"; break;
    }

    const char* LambdaCtrlMode = "DEFAULT_LDA";
    if (ctx->lambda_file)
        LambdaCtrlMode = "LOAD_LDA";

    const char* PrefetchBuffer = "DISABLE";
    switch (ctx->prefetch_buffer) {
        case 0: PrefetchBuffer = "DISABLE"; break;
        case 1: PrefetchBuffer = "ENABLE"; break;
    }

    sprintf (ctx->enc_options, "[INPUT]\n"
            "Width = %d\n"
            "Height = %d\n"
            "[RATE_CONTROL]\n"
            "RateCtrlMode = %s\n"
            "FrameRate = %s\n"
            "BitRate = %ld\n"
            "MaxBitRate = %ld\n"
            "SliceQP = %s\n"
            "MaxQP = %d\n"
            "MinQP = %d\n"
            "CPBSize = %f\n"
            "InitialDelay = %f\n"
            "[GOP]\n"
            "GopCtrlMode = %s\n"
            "Gop.GdrMode = %s\n"
            "Gop.Length = %d\n"
            "Gop.NumB = %d\n"
            "Gop.FreqIDR = %d\n"
            "[SETTINGS]\n"
            "Profile = %s\n"
            "Level = %s\n"
            "ChromaMode = CHROMA_4_2_0\n"
            "BitDepth = 8\n"
            "NumSlices = %d\n"
            "QPCtrlMode = %s\n"
            "SliceSize = %d\n"
            "EnableFillerData = %s\n"
            "AspectRatio = %s\n"
            "ColourDescription = %s\n"
            "ScalingList = %s\n"
            "EntropyMode = %s\n"
            "LoopFilter = %s\n"
            "ConstrainedIntraPred = %s\n"
            "LambdaCtrlMode = %s\n"
            "CacheLevel2 = %s\n"
            "NumCore = %d\n",
            avctx->width, avctx->height, RateCtrlMode, FrameRate, avctx->bit_rate / 1000,
            ctx->max_bitrate / 1000, SliceQP, ctx->max_qp, ctx->min_qp, ctx->cpb_size,
            ctx->initial_delay, GopCtrlMode, GDRMode, avctx->gop_size, ctx->b_frames,
            ctx->periodicity_idr, Profile, Level, ctx->num_slices, QPCtrlMode, ctx->slice_size,
            FillerData, AspectRatio, ColorSpace, ScalingList, EntropyMode, LoopFilter,
            ConstIntraPred, LambdaCtrlMode, PrefetchBuffer, ctx->cores);
}

static void fill_options_file_hevc (AVCodecContext *avctx)
{
    mpsoc_vcu_enc_ctx *ctx = avctx->priv_data;

    const char* RateCtrlMode = "CONST_QP";
    switch (ctx->control_rate) {
        case 0: RateCtrlMode = "CONST_QP"; break;
        case 1: RateCtrlMode = "CBR"; break;
        case 2: RateCtrlMode = "VBR"; break;
        case 3: RateCtrlMode = "LOW_LATENCY"; break;
        case 4: RateCtrlMode = "CAPPED_VBR"; break;
        case 5: RateCtrlMode = "BYPASS"; break;
    }

    char FrameRate[16];
    sprintf(FrameRate, "%llu/%llu", avctx->framerate.num, avctx->framerate.den);

    char SliceQP[8];
    if (ctx->slice_qp == -1)
        strcpy (SliceQP, "AUTO");
    else
        sprintf(SliceQP, "%d", ctx->slice_qp);

    const char* GopCtrlMode = "DEFAULT_GOP";
    switch (ctx->gop_mode) {
        case 0: GopCtrlMode = "DEFAULT_GOP"; break;
        case 1: GopCtrlMode = "PYRAMIDAL_GOP"; break;
        case 2: GopCtrlMode = "ADAPTIVE_GOP"; break;
        case 3: GopCtrlMode = "BYPASS"; break;
        case 4: GopCtrlMode = "LOW_DELAY_P"; break;
        case 5: GopCtrlMode = "LOW_DELAY_B"; break;
    }

    const char* GDRMode = "DISABLE";
    switch (ctx->gdr_mode) {
        case 0: GDRMode = "DISABLE"; break;
        case 1: GDRMode = "GDR_VERTICAL"; break;
        case 2: GDRMode = "GDR_HORIZONTAL"; break;
    }

    const char* Profile = "HEVC_MAIN";
    switch (ctx->profile) {
        case 0: Profile = "HEVC_MAIN"; break;
        case 1: Profile = "HEVC_MAIN_INTRA"; break;
        case 2: Profile = "HEVC_MAIN10"; break;
        case 3: Profile = "HEVC_MAIN10_INTRA"; break;
        case 4: Profile = "HEVC_MAIN_422"; break;
        case 5: Profile = "HEVC_MAIN_422_10"; break;
        case 6: Profile = "HEVC_MAIN_422_10_INTRA"; break;
    }

    const char* Level = "1";
    switch (ctx->level) {
        case 10: Level = "1"; break;
        case 20: Level = "2"; break;
        case 21: Level = "2.1"; break;
        case 30: Level = "3"; break;
        case 31: Level = "3.1"; break;
        case 40: Level = "4"; break;
        case 41: Level = "4.1"; break;
        case 50: Level = "5"; break;
        case 51: Level = "5.1"; break;
    }

    const char* Tier = "MAIN_TIER";
    switch (ctx->tier) {
        case 0: Tier = "MAIN_TIER"; break;
        case 1: Tier = "HIGH_TIER"; break;
    }

    const char* QPCtrlMode = "UNIFORM_QP";
    switch (ctx->qp_mode) {
        case 0: QPCtrlMode = "UNIFORM_QP"; break;
        case 1: QPCtrlMode = "AUTO_QP"; break;
        case 2: QPCtrlMode = "LOAD_QP"; break;
        case 3: QPCtrlMode = "LOAD_QP | RELATIVE_QP"; break;
    }

    const char* DependentSlice = "FALSE";
    switch (ctx->dependent_slice) {
        case 0: DependentSlice = "FALSE"; break;
        case 1: DependentSlice = "TRUE"; break;
    }

    const char* FillerData = "ENABLE";
    switch (ctx->filler_data) {
        case 0: FillerData = "DISABLE"; break;
        case 1: FillerData = "ENABLE"; break;
    }

    const char* AspectRatio = "ASPECT_RATIO_AUTO";
    switch (ctx->aspect_ratio) {
        case 0: AspectRatio = "ASPECT_RATIO_AUTO"; break;
        case 1: AspectRatio = "ASPECT_RATIO_4_3"; break;
        case 2: AspectRatio = "ASPECT_RATIO_16_9"; break;
        case 3: AspectRatio = "ASPECT_RATIO_NONE"; break;
    }

    const char* ColorSpace = "COLOUR_DESC_UNSPECIFIED";
    switch (avctx->colorspace) {
        case AVCOL_SPC_BT709: ColorSpace = "COLOUR_DESC_BT_709"; break;
        case AVCOL_SPC_UNSPECIFIED: ColorSpace = "COLOUR_DESC_UNSPECIFIED"; break;
        case AVCOL_SPC_RESERVED: ColorSpace = "COLOUR_DESC_RESERVED"; break;
        case AVCOL_SPC_BT470BG: ColorSpace = "COLOUR_DESC_BT_470_NTSC"; break;
        case AVCOL_SPC_SMPTE170M: ColorSpace = "COLOUR_DESC_BT_601_PAL"; break;
        case AVCOL_SPC_SMPTE240M: ColorSpace = "COLOUR_DESC_BT_601_NTSC"; break;
        case AVCOL_SPC_BT2020_NCL: ColorSpace = "COLOUR_DESC_BT_2020"; break;
        case AVCOL_SPC_BT2020_CL: ColorSpace = "COLOUR_DESC_BT_2020"; break;
    }

    const char* ScalingList = "FLAT";
    switch (ctx->scaling_list) {
        case 0: ScalingList = "FLAT"; break;
        case 1: ScalingList = "DEFAULT"; break;
    }

    const char* LoopFilter = "ENABLE";
    switch (ctx->loop_filter) {
        case 0: LoopFilter = "DISABLE"; break;
        case 1: LoopFilter = "ENABLE"; break;
    }

    const char* ConstIntraPred = "ENABLE";
    switch (ctx->constrained_intra_pred) {
        case 0: ConstIntraPred = "DISABLE"; break;
        case 1: ConstIntraPred = "ENABLE"; break;
    }

    const char* LambdaCtrlMode = "DEFAULT_LDA";
    if (ctx->lambda_file)
        LambdaCtrlMode = "LOAD_LDA";

    const char* PrefetchBuffer = "DISABLE";
    switch (ctx->prefetch_buffer) {
        case 0: PrefetchBuffer = "DISABLE"; break;
        case 1: PrefetchBuffer = "ENABLE"; break;
    }

    sprintf (ctx->enc_options, "[INPUT]\n"
            "Width = %d\n"
            "Height = %d\n"
            "[RATE_CONTROL]\n"
            "RateCtrlMode = %s\n"
            "FrameRate = %s\n"
            "BitRate = %ld\n"
            "MaxBitRate = %ld\n"
            "SliceQP = %s\n"
            "MaxQP = %d\n"
            "MinQP = %d\n"
            "CPBSize = %f\n"
            "InitialDelay = %f\n"
            "[GOP]\n"
            "GopCtrlMode = %s\n"
            "Gop.GdrMode = %s\n"
            "Gop.Length = %d\n"
            "Gop.NumB = %d\n"
            "Gop.FreqIDR = %d\n"
            "[SETTINGS]\n"
            "Profile = %s\n"
            "Level = %s\n"
            "Tier = %s\n"
            "ChromaMode = CHROMA_4_2_0\n"
            "BitDepth = 8\n"
            "NumSlices = %d\n"
            "QPCtrlMode = %s\n"
            "SliceSize = %d\n"
            "DependentSlice = %s\n"
            "EnableFillerData = %s\n"
            "AspectRatio = %s\n"
            "ColourDescription = %s\n"
            "ScalingList = %s\n"
            "LoopFilter = %s\n"
            "ConstrainedIntraPred = %s\n"
            "LambdaCtrlMode = %s\n"
            "CacheLevel2 = %s\n"
            "NumCore = %d\n",
            avctx->width, avctx->height, RateCtrlMode, FrameRate, avctx->bit_rate / 1000,
            ctx->max_bitrate / 1000, SliceQP, ctx->max_qp, ctx->min_qp, ctx->cpb_size,
            ctx->initial_delay, GopCtrlMode, GDRMode, avctx->gop_size, ctx->b_frames,
            ctx->periodicity_idr, Profile, Level, Tier, ctx->num_slices, QPCtrlMode,
            ctx->slice_size, DependentSlice, FillerData, AspectRatio, ColorSpace, ScalingList,
            LoopFilter, ConstIntraPred, LambdaCtrlMode, PrefetchBuffer, ctx->cores);
}

#define MIN_LOOKAHEAD_DEPTH	(1)
#define MAX_LOOKAHEAD_DEPTH	(30)

static void _allocate_xrm_enc_cu(mpsoc_vcu_enc_ctx *ctx, XmaEncoderProperties *enc_props)
{


    int ret =-1;
    char pluginName[XRM_MAX_NAME_LEN];

    if (getenv("XRM_RESERVE_ID"))
    {
       int xrm_reserve_id = atoi(getenv("XRM_RESERVE_ID"));
       //create XRM local context
       ctx->xrm_ctx = (xrmContext *)xrmCreateContext(XRM_API_VERSION_1);
       if (ctx->xrm_ctx == NULL)
       {
          printf("create local XRM context failed\n");
          return;
       }

       //XRM encoder plugin load calculation
       int func_id = 0, enc_load=0;
       xrmPluginFuncParam param;
       memset(&param, 0, sizeof(xrmPluginFuncParam));
       void *handle;
       void (*convertXmaPropsToJson)(void* props, char* funcName, char* jsonJob);

       handle = dlopen("/opt/xilinx/xrm/plugin/libxmaPropsTOjson.so", RTLD_NOW );

       convertXmaPropsToJson = dlsym(handle, "convertXmaPropsToJson");
       (*convertXmaPropsToJson) (enc_props, "ENCODER",param.input);
       dlclose(handle);

       strcpy(pluginName, "xrmU30EncPlugin");

       if (xrmExecPluginFunc(ctx->xrm_ctx, pluginName, func_id, &param) != XRM_SUCCESS)
           printf("enc plugin function %d, fail to run the function\n", func_id);
       else
       {
           enc_load = atoi((char*)(strtok(param.output, " ")));
           //printf("enc plugin function %d, success to run the function, input: %s, output: %s enc_load=%d\n", func_id, param.input, param.output, enc_load);
       }

       //XRM encoder allocation
       xrmCuListProperty encode_cu_list_prop;

       memset(&encode_cu_list_prop, 0, sizeof(xrmCuListProperty));
       memset(&ctx->encode_cu_list_res, 0, sizeof(xrmCuListResource));

       encode_cu_list_prop.cuNum = 2;
       strcpy(encode_cu_list_prop.cuProps[0].kernelName, "encoder");
       strcpy(encode_cu_list_prop.cuProps[0].kernelAlias, "ENCODER_MPSOC");
       encode_cu_list_prop.cuProps[0].devExcl = false;
       encode_cu_list_prop.cuProps[0].requestLoad = enc_load;
       encode_cu_list_prop.cuProps[0].poolId = xrm_reserve_id;

       strcpy(encode_cu_list_prop.cuProps[1].kernelName, "kernel_vcu_encoder");
       encode_cu_list_prop.cuProps[1].devExcl = false;
       encode_cu_list_prop.cuProps[1].requestLoad = 100;
       encode_cu_list_prop.cuProps[1].poolId = xrm_reserve_id;

       //printf("xrm enc list allocation %d\n",enc_load);
       ret = xrmCuListAlloc(ctx->xrm_ctx, &encode_cu_list_prop, &ctx->encode_cu_list_res);

       if (ret != 0)
       {
          printf("xrm_allocation: fail to allocate encoder cu  from reserve\n");
          return;
       }
       else
       {
          ctx->encode_res_inuse = 1;
       }

       //Set XMA plugin SO and device index
       enc_props->plugin_lib = ctx->encode_cu_list_res.cuResources[0].kernelPluginFileName;
       enc_props->dev_index = ctx->encode_cu_list_res.cuResources[0].deviceId;
       enc_props->ddr_bank_index = -1;//XMA to select the ddr bank based on xclbin meta data
       enc_props->cu_index = ctx->encode_cu_list_res.cuResources[1].cuId;
       enc_props->channel_id = ctx->encode_cu_list_res.cuResources[1].channelId;

       //printf("enc xrm out %s %d %d %d %d\n",enc_props->plugin_lib, enc_props->dev_index,enc_props->ddr_bank_index, enc_props->cu_index, enc_props->channel_id );
    }
    else
    {
       //Set XMA plugin SO and device index
       enc_props->plugin_lib = "/opt/xilinx/xma_plugins/libvcu-xma-enc-plg.so";
       enc_props->dev_index = 0;
       enc_props->ddr_bank_index = -1;//XMA to select the ddr bank based on xclbin meta data
       enc_props->cu_index = global_enc_chan_id;
       enc_props->channel_id = 0;
       global_enc_chan_id++;
    }
}

static av_cold int mpsoc_vcu_encode_init(AVCodecContext *avctx)
{
    XmaEncoderProperties enc_props;
    mpsoc_vcu_enc_ctx *ctx = avctx->priv_data;
    int out_params_size = 0;
    int ret = 0;
    char* out_params =NULL;

    if (ctx->lookahead_depth > avctx->gop_size) {
        printf("Error : mpsoc_vcu_encode_frame : Invalid arguments. gop size(%d) should be grater than lookahead_depth(%d)\n", avctx->gop_size, ctx->lookahead_depth);
        return AVERROR(EINVAL);
    }

    if (avctx->codec_id == AV_CODEC_ID_H264) fill_options_file_h264 (avctx);
    else if (avctx->codec_id == AV_CODEC_ID_HEVC) fill_options_file_hevc (avctx);
    else {
        printf("Unknown codec id!\n");
        return AVERROR_ENCODER_NOT_FOUND;
    }
    enc_props.hwencoder_type = XMA_MULTI_ENCODER_TYPE;
    strcpy(enc_props.hwvendor_string, "MPSoC");

    enc_props.format = XMA_VCU_NV12_FMT_TYPE;
    enc_props.bits_per_pixel  = 8;
    enc_props.param_cnt = 0;
    enc_props.params = ctx->enc_params;
    enc_props.width = avctx->width;
    enc_props.height = avctx->height;
    enc_props.rc_mode =  ctx->rate_control_mode;

    if (ctx->lookahead_depth)
    {
        //Validate input resolution against MAX supported
        if ((enc_props.width > MAX_INPUT_WIDTH) || //for landscape use-case
            (enc_props.height > MAX_INPUT_WIDTH) || //for portrait use-case
            ((enc_props.width * enc_props.height) > MAX_INPUT_PIXELS)) {
            av_log (ctx, AV_LOG_ERROR, "\n Input to encoder for look ahead usecase %4dx%4d exceeds max supported resolution %4dx%4d (or %4dx%4d portrait mode)\n\n",
                    enc_props.width ,enc_props.height, MAX_INPUT_WIDTH, MAX_INPUT_HEIGHT, MAX_INPUT_HEIGHT, MAX_INPUT_WIDTH);
            return -1;
        }
    }

    switch(enc_props.rc_mode) {
      case 0 : printf("Custom Rate Control Mode is Disabled\n");
               break;
      case 1 : if (ctx->lookahead_depth < MIN_LOOKAHEAD_DEPTH ||
                   ctx->lookahead_depth > MAX_LOOKAHEAD_DEPTH) {
                 printf("Error: Provided LA Depth %d is invalid !\n", ctx->lookahead_depth);
                 printf("If rate-control-mode = 1 the lookahead_depth must lie between %d - %d\n",
                        MIN_LOOKAHEAD_DEPTH, MAX_LOOKAHEAD_DEPTH);
                 return AVERROR(EINVAL);
               } else {
                 enc_props.lookahead_depth = ctx->lookahead_depth;
               }
               printf("#### Custom Rate Control Mode is Enabled with LA Depth = %d ####\n", enc_props.lookahead_depth);
               break;
      default: enc_props.rc_mode = 0;
               printf("Rate Control Mode is default\n");
               break;
    }

    enc_props.framerate.numerator   = avctx->framerate.num;
    enc_props.framerate.denominator = avctx->framerate.den;
    ctx->frame.frame_props.format = XMA_VCU_NV12_FMT_TYPE;
    ctx->frame.frame_props.width = FFALIGN(avctx->width, 256);
    ctx->frame.frame_props.height = FFALIGN(avctx->height, 64);
    ctx->frame.frame_props.bits_per_pixel = 8;

    const char* enc_options = ctx->enc_options;
    if (avctx->pix_fmt == AV_PIX_FMT_VCU_NV12)
        ctx->frame.data[0].buffer_type = XMA_DEVICE_BUFFER_TYPE;
    else if (avctx->pix_fmt == AV_PIX_FMT_NV12)
    	ctx->frame.data[0].buffer_type = XMA_HOST_BUFFER_TYPE;
    ctx->enc_params[enc_props.param_cnt].name   = "enc_options";
    ctx->enc_params[enc_props.param_cnt].type   = XMA_STRING;
    ctx->enc_params[enc_props.param_cnt].length = strlen(ctx->enc_options);
    ctx->enc_params[enc_props.param_cnt].value  = &(enc_options);
    enc_props.param_cnt++;

    ctx->enc_params[enc_props.param_cnt].name   = "latency_logging";
    ctx->enc_params[enc_props.param_cnt].type   = XMA_UINT32;
    ctx->enc_params[enc_props.param_cnt].length = sizeof(ctx->latency_logging);
    ctx->enc_params[enc_props.param_cnt].value  = &(ctx->latency_logging);
    enc_props.param_cnt++;

    /* optional parameter */
    if (ctx->lambda_file) {
        ctx->enc_params[enc_props.param_cnt].name   = "lambda_file";
        ctx->enc_params[enc_props.param_cnt].type   = XMA_STRING;
        ctx->enc_params[enc_props.param_cnt].length = strlen(ctx->lambda_file);
        ctx->enc_params[enc_props.param_cnt].value  = &(ctx->lambda_file);
        enc_props.param_cnt++;
    }
    ctx->sent_flush = false;

    ctx->la = NULL;

    if(init_la(avctx)) {
        printf("Error: Unable to init_la Invalid params\n");
        return AVERROR(EINVAL);
    }

    /*----------------------------------------------------
      Allocate encoder resource from XRM reserved resource
      ----------------------------------------------------*/
    ctx->encode_res_inuse = 0;
    _allocate_xrm_enc_cu(ctx, &enc_props);

    ctx->enc_session = xma_enc_session_create(&enc_props);
    if (!ctx->enc_session) {
        printf("ERROR: Unable to allocate MPSoC encoder session\n");
        return AVERROR_EXTERNAL;
    }
    ctx->pts_0 = AV_NOPTS_VALUE;
    ctx->pts_1 = AV_NOPTS_VALUE;
    ctx->is_first_outframe = 1;

    // TODO: find a proper way to find pts_queue size
    ctx->pts_queue = av_fifo_alloc(64 * sizeof(int64_t));
    if (!ctx->pts_queue)
        return AVERROR(ENOMEM);

    return 0;
}

static void vcu_enc_free_out_buffer(void *opaque, uint8_t *data)
{
    /*do nothing, if this CB is not provided, ffmpeg tries to free xrt buffer */
}

int vcu_alloc_ff_packet(mpsoc_vcu_enc_ctx *ctx, AVPacket *pkt)
{
    pkt->buf = av_buffer_create(ctx->xma_buffer.data.buffer, pkt->size, vcu_enc_free_out_buffer,  NULL,
                                AV_GET_BUFFER_FLAG_REF);
    if (!pkt->buf)
        return AVERROR(ENOMEM);
    pkt->size = pkt->size;

    pkt->data = ctx->xma_buffer.data.buffer;
    pkt->size = pkt->size;
    return 0;
}

static int mpsoc_vcu_encode_frame(AVCodecContext *avctx, AVPacket *pkt, const AVFrame *pic, int *got_packet)
{
    mpsoc_vcu_enc_ctx *ctx = avctx->priv_data;
    int recv_size, ret;
    XmaFrame *la_in_frame = NULL;
    XmaFrame *enc_in_frame = NULL;
    *got_packet = 0;
    recv_size = 0;

    if (pic && pic->data) {
        if (av_fifo_size(ctx->laQ) < sizeof(XmaFrame *)) {
            printf("Error : insufficient LA frames\n");
            return AVERROR_EXTERNAL;
        }
        av_fifo_generic_read(ctx->laQ, &la_in_frame, sizeof(XmaFrame *), NULL);
        if (avctx->pix_fmt == AV_PIX_FMT_VCU_NV12) {
            XmaSideDataHandle *side_data = la_in_frame->side_data;
            memcpy (la_in_frame, pic->data[0], sizeof (XmaFrame));
            xvbm_buffer_refcnt_inc(la_in_frame->data[0].buffer);
            la_in_frame->pts = pic->pts;
            la_in_frame->side_data = side_data;

            mpsoc_vcu_encode_queue_pts(ctx->pts_queue, la_in_frame->pts);
        } else {
            //@TODO copy/resuse the YUV data for other formats
            uint8_t *dst;
            uint8_t *src;
            uint32_t dst_stride;
            uint32_t src_stride;
            uint32_t i;
            uint32_t bf_width = FFALIGN(avctx->width, 256);
            uint32_t bf_height = FFALIGN(avctx->height, 64);
            la_in_frame->data[0].refcount = 1;
            la_in_frame->data[0].buffer_type = XMA_HOST_BUFFER_TYPE;
            la_in_frame->data[0].is_clone = false;
            la_in_frame->data[0].buffer = calloc(1, bf_width * bf_height);
            la_in_frame->data[1].refcount = 1;
            la_in_frame->data[1].buffer_type = XMA_HOST_BUFFER_TYPE;
            la_in_frame->data[1].is_clone = false;
            la_in_frame->data[1].buffer = calloc(1, bf_width * bf_height);
            la_in_frame->side_data = NULL;
            dst = la_in_frame->data[0].buffer;
            src = pic->data[0];
            dst_stride = bf_width;
            src_stride = pic->linesize[0];
            for (i = 0; i < pic->height; i++) {
                memcpy (dst, src, pic->width);
                dst += dst_stride;
                src += src_stride;
            }
            dst = la_in_frame->data[1].buffer;
            src = pic->data[1];
            src_stride = pic->linesize[1];
            for (i = 0; i < pic->height / 2; i++) {
                memcpy (dst, src, pic->width);
                dst += dst_stride;
                src += src_stride;
            }
            la_in_frame->pts = pic->pts;
            mpsoc_vcu_encode_queue_pts(ctx->pts_queue, la_in_frame->pts);
        }

        if (ctx->pts_0 == AV_NOPTS_VALUE)
            ctx->pts_0 = la_in_frame->pts;
        else if (ctx->pts_1 == AV_NOPTS_VALUE)
            ctx->pts_1 = la_in_frame->pts;
    }

    ret = xlnx_la_send_recv_frame(ctx->la, la_in_frame, &enc_in_frame);
    if (ret <= XMA_ERROR) {
    	if (avctx->pix_fmt == AV_PIX_FMT_VCU_NV12) {
            XvbmBufferHandle handle = (XvbmBufferHandle)(la_in_frame->data[0].buffer);
            if (handle) {
                xvbm_buffer_pool_entry_free(handle);
            }
    	}
        xma_frame_clear_all_side_data(la_in_frame);
        av_free(la_in_frame);
        return AVERROR(EIO);
    } else if ((ret == XMA_SEND_MORE_DATA) && (la_in_frame != NULL)) {
        return 0;
    }
    if (enc_in_frame) {
        ret = xma_enc_session_send_frame(ctx->enc_session, enc_in_frame);
        av_fifo_generic_write(ctx->laQ, &enc_in_frame, sizeof(XmaFrame *), NULL);
        if (ret == XMA_SUCCESS) {
            while (1) {
                ret = xma_enc_session_recv_data(ctx->enc_session, &(ctx->xma_buffer), &recv_size);
                if (ret == XMA_SUCCESS) {
                    /* valid data received */
                    *got_packet = 1;
                    pkt->size = recv_size;
                    vcu_alloc_ff_packet(ctx, pkt);
                    pkt->pts = ctx->xma_buffer.pts;
                    mpsoc_vcu_encode_prepare_out_timestamp (avctx, pkt);
                    pkt->flags |= ((avctx->codec_id == AV_CODEC_ID_H264) ? mpsoc_encode_is_h264_idr (pkt) : mpsoc_encode_is_hevc_idr (pkt)) ? AV_PKT_FLAG_KEY : 0;
                    break;
                } else if (ret == XMA_TRY_AGAIN) {
                    if (pic && pic->data) {
                        /* vcu not ready with an output buffer */
                        *got_packet = 0;
                        return 0;
                    }
                } else {
                    /* vcu not ready with an output buffer */
                    *got_packet = 0;
                    return 0;
                }
            }
        } else {
            /* send raw data failed */
            *got_packet = 0;
            printf("Error : send raw data failed\n");
            return AVERROR(EIO);
        }
    } else { /* end of input data */
        ctx->frame.is_last_frame = 1;
        if (ctx->sent_flush == false) {
            ctx->sent_flush = true;
            ctx->frame.pts = -1;
            ret = xma_enc_session_send_frame(ctx->enc_session, &(ctx->frame));
            if (ret == XMA_SUCCESS) {
                ret = xma_enc_session_recv_data(ctx->enc_session, &(ctx->xma_buffer), &recv_size);
                if (ret == XMA_SUCCESS) {
                    *got_packet = 1;
                    pkt->size = recv_size;
                    vcu_alloc_ff_packet(ctx, pkt);
                    pkt->pts = ctx->xma_buffer.pts;
                    mpsoc_vcu_encode_prepare_out_timestamp (avctx, pkt);
                    pkt->flags |= ((avctx->codec_id == AV_CODEC_ID_H264) ? mpsoc_encode_is_h264_idr (pkt) : mpsoc_encode_is_hevc_idr (pkt)) ? AV_PKT_FLAG_KEY : 0;
                } else {
                    *got_packet = 0;
                }
            }
        } else {
            ret = xma_enc_session_recv_data(ctx->enc_session, &(ctx->xma_buffer), &recv_size);
            if (ret == XMA_SUCCESS) {
                pkt->size = recv_size;
                vcu_alloc_ff_packet(ctx, pkt);
                pkt->pts = ctx->xma_buffer.pts;
                mpsoc_vcu_encode_prepare_out_timestamp (avctx, pkt);
                pkt->flags |= ((avctx->codec_id == AV_CODEC_ID_H264) ? mpsoc_encode_is_h264_idr (pkt) : mpsoc_encode_is_hevc_idr (pkt)) ? AV_PKT_FLAG_KEY : 0;
                *got_packet = 1;
            } else {
                *got_packet = 0;
            }
        }
    }

    return 0;
}

static const AVCodecDefault mpsoc_defaults[] = {
    { "b", "5M" },
    { "g", "120" },
    { NULL },
};

static const AVClass mpsoc_h264_class = {
    .class_name = "MPSOC VCU H264 encoder",
    .item_name = av_default_item_name,
    .option = h264Options,
    .version = LIBAVUTIL_VERSION_INT,
};

AVCodec ff_h264_vcu_mpsoc_encoder = {
    .name = "mpsoc_vcu_h264",
    .long_name = NULL_IF_CONFIG_SMALL("MPSOC H.264 Encoder"),
    .type = AVMEDIA_TYPE_VIDEO,
    .id = AV_CODEC_ID_H264,
    .init = mpsoc_vcu_encode_init,
    .encode2 = mpsoc_vcu_encode_frame,
    .close = mpsoc_vcu_encode_close,
    .priv_data_size = sizeof(mpsoc_vcu_enc_ctx),
    .priv_class = &mpsoc_h264_class,
    .defaults = mpsoc_defaults,
    .pix_fmts = (const enum AVPixelFormat[]) { AV_PIX_FMT_VCU_NV12,
                                               AV_PIX_FMT_NV12,
                                               AV_PIX_FMT_NONE },
    .capabilities = AV_CODEC_CAP_DELAY | AV_CODEC_CAP_AUTO_THREADS,
};

static const AVClass mpsoc_hevc_vcu_class = {
    .class_name = "MPSOC VCU HEVC encoder",
    .item_name = av_default_item_name,
    .option = hevcOptions,
    .version = LIBAVUTIL_VERSION_INT,
};

AVCodec ff_hevc_vcu_mpsoc_encoder = {
    .name = "mpsoc_vcu_hevc",
    .long_name = NULL_IF_CONFIG_SMALL("MPSOC VCU HEVC Encoder"),
    .type  = AVMEDIA_TYPE_VIDEO,
    .id = AV_CODEC_ID_HEVC,
    .init = mpsoc_vcu_encode_init,
    .encode2 = mpsoc_vcu_encode_frame,
    .close = mpsoc_vcu_encode_close,
    .priv_data_size = sizeof(mpsoc_vcu_enc_ctx),
    .priv_class = &mpsoc_hevc_vcu_class,
    .defaults = mpsoc_defaults,
    .pix_fmts = (const enum AVPixelFormat[]) { AV_PIX_FMT_VCU_NV12,
                                               AV_PIX_FMT_NV12,
                                               AV_PIX_FMT_NONE },
    .capabilities = AV_CODEC_CAP_DELAY | AV_CODEC_CAP_AUTO_THREADS | AV_CODEC_CAP_AVOID_PROBING,
};
