/*
* Copyright (c) 2018 Xilinx Inc
*
* This file is part of FFmpeg.
*
* FFmpeg is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* FFmpeg is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with FFmpeg; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
*/

#include <stdlib.h>
#include <string.h>
#include "libavutil/fifo.h"
#include "xlnx_lookahead.h"
#include "xvbm.h"
#include <xrm.h>
#include <dlfcn.h>
#include "../xmaPropsTOjson.h"

//From #include "xlnx_la_plg_ext.h"
#define XLNX_LA_PLG_NUM_EXT_PARAMS 10

typedef enum
{
    EParamIntraPeriod,
    EParamLADepth,
    EParamEnableZeroCopy,
    EParamSpatialAQMode,
    EParamTemporalAQMode,
    EParamRateControlMode,
    EParamSpatialAQGain,
    EParamNumBFrames,
    EParamCodecType,
	EParamLatencyLogging
} xlnx_la_ext_params_t;

static const char *XLNX_LA_EXT_PARAMS[] = {
    "ip",
    "lookahead_depth",
    "enable_zerocopy",
    "spatial_aq_mode",
    "temporal_aq_mode",
    "rate_control_mode",
    "spatial_aq_gain",
    "num_b_frames",
    "codec_type",
	"latency_logging"
};

///////////////////////////////////////////////////////////////////////////////

#define SCLEVEL1 2
#define XLNX_MAX_LOOKAHEAD_DEPTH 20
#define XLNX_ALIGN(x,LINE_SIZE) (((((size_t)x) + ((size_t)LINE_SIZE - 1)) & (~((size_t)LINE_SIZE - 1))))

static const char *XLNX_LOOKAHEAD_NAME = "xlnx_lookahead";

#define XLNX_LA_LOG(LOG_TYPE, ...)                               \
    do {                                                         \
        xma_logmsg(LOG_TYPE, XLNX_LOOKAHEAD_NAME, __VA_ARGS__);  \
    } while (0)

typedef struct xlnx_la_ctx
{
    XmaFilterSession *filter_session;
    XmaFrame *qp_frame;
    AVFifoBuffer *in_buf_Q;
    uint8_t bypass;
    uint32_t enableZeroCopy;
    uint32_t lookahead_depth;
    uint32_t spatial_aq_mode;
    uint32_t temporal_aq_mode;
    uint32_t rate_control_mode;
    uint32_t spatial_aq_gain;
    XmaFormatType fmt_type;
    int32_t num_planes;
    xlnx_codec_type_t codec_type;
    XmaParameter extn_params[XLNX_LA_PLG_NUM_EXT_PARAMS];
    xrmContext        *xrm_ctx;
    xrmCuListResource  lookahead_cu_list_res;
    int                lookahead_res_inuse;
} xlnx_la_ctx;

static int32_t
get_num_video_planes(XmaFormatType format)
{
    /* multi scaler supports max 2 planes till v2019.1 */
    switch (format) {
        case XMA_RGB888_FMT_TYPE: /* BGR */
            return 1;
        case XMA_YUV420_FMT_TYPE: /* NV12 */
            return 2;
        case XMA_VCU_NV12_FMT_TYPE: /* VCU_NV12 */
            return 1;
        default:
            XLNX_LA_LOG(XMA_ERROR_LOG, "Unsupported format...");
            return -1;
    }
}

static void free_frame(XmaFrame *in_frame)
{
    XvbmBufferHandle handle;
    int32_t num_planes;
    if (in_frame == NULL) {
        return;
    }
    if (in_frame->data[0].buffer_type == XMA_DEVICE_BUFFER_TYPE) {
        handle = (XvbmBufferHandle)(in_frame->data[0].buffer);
        if (handle) {
            xvbm_buffer_pool_entry_free(handle);
        }
    } else {
        num_planes = xma_frame_planes_get(&in_frame->frame_props);

        for (int32_t i = 0; i < num_planes; i++) {
            in_frame->data[i].refcount--;
        }

        if (in_frame->data[0].refcount > 0) {
            return;
        }

        for (int32_t i = 0; i < num_planes && !in_frame->data[i].is_clone; i++) {
            free(in_frame->data[i].buffer);
        }
    }
    xma_frame_clear_all_side_data(in_frame);
}

static int32_t free_res(xlnx_la_ctx *la_ctx)
{
    XmaFrame *in_frame = NULL;

    if (!la_ctx) {
        XLNX_LA_LOG(XMA_ERROR_LOG, "free_res : free_res la_ctx = NULL\n");
        return XMA_ERROR;
    }

    if (getenv("XRM_RESERVE_ID")) {
        int xrm_reserve_id = atoi(getenv("XRM_RESERVE_ID"));

        //XRM lookahead de-allocation
        if (la_ctx->lookahead_res_inuse ==1) {
            if (!(xrmCuListRelease(la_ctx->xrm_ctx, &la_ctx->lookahead_cu_list_res))) {
                printf("fail to release lookahead HW cu\n");
            }
        if (xrmDestroyContext(la_ctx->xrm_ctx) != XRM_SUCCESS)
            printf("XRM : destroy context failed\n");

        }
    }

    // Close lookahead session
    if (la_ctx->filter_session) {
        xma_filter_session_destroy(la_ctx->filter_session);
        la_ctx->filter_session = NULL;
    }

    if (la_ctx->qp_frame) {
        free_frame(la_ctx->qp_frame);
        av_free(la_ctx->qp_frame);
        la_ctx->qp_frame = NULL;
    }
    if (la_ctx->in_buf_Q) {
        while (av_fifo_size(la_ctx->in_buf_Q) > sizeof(XmaFrame *)) {
            av_fifo_generic_read(la_ctx->in_buf_Q, &in_frame, sizeof(XmaFrame *), NULL);
            free_frame(in_frame);
            av_free(in_frame);
        }
        av_fifo_free(la_ctx->in_buf_Q);
        la_ctx->in_buf_Q = NULL;
    }

    return XMA_SUCCESS;
}

static void _allocate_xrm_la_cu(xlnx_la_ctx *ctx,
                                XmaFilterProperties *filter_props)
{


    int ret =-1;
    char pluginName[XRM_MAX_NAME_LEN];

    if (getenv("XRM_RESERVE_ID")) {
        int xrm_reserve_id = atoi(getenv("XRM_RESERVE_ID"));
        //create XRM local context
        ctx->xrm_ctx = (xrmContext *)xrmCreateContext(XRM_API_VERSION_1);
        if (ctx->xrm_ctx == NULL) {
            printf("create local XRM context failed\n");
            return;
        }

        //XRM encoder plugin load calculation
        int func_id = 0, la_load=0, skip_value=0;
        xrmPluginFuncParam param;
        memset(&param, 0, sizeof(xrmPluginFuncParam));
        void *handle;
        void (*convertXmaPropsToJson)(void *props, char *funcName, char *jsonJob);

        handle = dlopen("/opt/xilinx/xrm/plugin/libxmaPropsTOjson.so", RTLD_NOW );

        convertXmaPropsToJson = dlsym(handle, "convertXmaPropsToJson");
        (*convertXmaPropsToJson) (filter_props, "LOOKAHEAD", param.input);
        dlclose(handle);

        strcpy(pluginName, "xrmU30EncPlugin");
        if (xrmExecPluginFunc(ctx->xrm_ctx, pluginName, func_id,
                              &param) != XRM_SUCCESS) {
            printf("lookahead plugin function %d, fail to run the function\n", func_id);
        } else {
            skip_value = atoi((char *)(strtok(param.output, " ")));
            skip_value = atoi((char *)(strtok(NULL, " ")));
            la_load = atoi((char *)(strtok(NULL, " ")));
            //printf("\n\n-******--lookahead plugin function %d, success to run the function, input: %s, output: %s la_load=%d\n\n\n", func_id, param.input, param.output, la_load);
        }

        //XRM lookahead allocation
        xrmCuListProperty lookahead_cu_list_prop;

        memset(&lookahead_cu_list_prop, 0, sizeof(xrmCuListProperty));
        memset(&ctx->lookahead_cu_list_res, 0, sizeof(xrmCuListResource));

        lookahead_cu_list_prop.cuNum = 1;
        strcpy(lookahead_cu_list_prop.cuProps[0].kernelName, "lookahead");
        strcpy(lookahead_cu_list_prop.cuProps[0].kernelAlias, "LOOKAHEAD_MPSOC");
        lookahead_cu_list_prop.cuProps[0].devExcl = false;
        lookahead_cu_list_prop.cuProps[0].requestLoad = la_load;
        lookahead_cu_list_prop.cuProps[0].poolId = xrm_reserve_id;

        ret = xrmCuListAlloc(ctx->xrm_ctx, &lookahead_cu_list_prop,
                             &ctx->lookahead_cu_list_res);

        if (ret != 0) {
            printf("xrm_allocation: fail to allocate lookahead cu  from reserve\n");
            return;
        } else {
            ctx->lookahead_res_inuse = 1;
        }

        //Set XMA plugin SO and device index
        filter_props->plugin_lib =
            ctx->lookahead_cu_list_res.cuResources[0].kernelPluginFileName;
        filter_props->dev_index = ctx->lookahead_cu_list_res.cuResources[0].deviceId;
        filter_props->ddr_bank_index =
            -1;//XMA to select the ddr bank based on xclbin meta data
        filter_props->cu_index = ctx->lookahead_cu_list_res.cuResources[0].cuId;
        filter_props->channel_id = ctx->lookahead_cu_list_res.cuResources[0].channelId;
    } else {
        //Set XMA plugin SO and device index
        filter_props->plugin_lib = "/opt/xilinx/xma_plugins/libxlookahead.so";
        filter_props->dev_index = 0;
        filter_props->ddr_bank_index =
            -1;//ALERT:-1;//XMA to select the ddr bank based on xclbin meta data
        filter_props->cu_index = 2;
        filter_props->channel_id = 0;
        //global_la_chan_id++;
    }
    //printf("lookahead xrm out %s %d %d %d %d\n",filter_props->plugin_lib, filter_props->dev_index,filter_props->ddr_bank_index, filter_props->cu_index, filter_props->channel_id );
}


xlnx_lookahead_t create_xlnx_la(xlnx_la_cfg_t *cfg)
{
    XmaFilterProperties filter_props;
    XmaFilterPortProperties *in_props;
    XmaFilterPortProperties *out_props;
    uint32_t param_cnt = 0;
    //XmaFrameProperties f_in_props;
    //XmaFrameProperties f_out_props;
    xlnx_la_ctx *la_ctx;
    XmaParameter *extn_params = NULL;

    if (!cfg) {
        XLNX_LA_LOG(XMA_ERROR_LOG, "No config received\n");
        return NULL;
    }
    if ((cfg->lookahead_depth == 0) && (cfg->temporal_aq_mode == 1)) {
        XLNX_LA_LOG(XMA_ERROR_LOG, "Invalid params: Lookahead = 0, temporal aq=%u\n",
                    cfg->temporal_aq_mode);
        return NULL;
    }


    la_ctx = calloc(1, sizeof(xlnx_la_ctx));
    if (!la_ctx) {
        XLNX_LA_LOG(XMA_ERROR_LOG, "OOM la_ctx\n");
        return NULL;
    }
    la_ctx->lookahead_depth = cfg->lookahead_depth;
    if (((cfg->lookahead_depth == 0) && (cfg->spatial_aq_mode == 0)) ||
            ((cfg->spatial_aq_mode == 0) && (cfg->temporal_aq_mode == 0) &&
             (cfg->rate_control_mode == 0))) {
        la_ctx->bypass = 1;
        return la_ctx;
    }
    la_ctx->spatial_aq_mode = cfg->spatial_aq_mode;
    la_ctx->temporal_aq_mode = cfg->temporal_aq_mode;
    la_ctx->spatial_aq_gain = cfg->spatial_aq_gain;
    la_ctx->enableZeroCopy = cfg->enableZeroCopy;
    la_ctx->fmt_type = cfg->fmt_type;
    la_ctx->num_planes = get_num_video_planes(la_ctx->fmt_type);
    la_ctx->rate_control_mode = cfg->rate_control_mode;
    la_ctx->bypass = 0;
    la_ctx->codec_type = cfg->codec_type;
    if (la_ctx->lookahead_depth > 0) {
        la_ctx->in_buf_Q = av_fifo_alloc((XLNX_MAX_LOOKAHEAD_DEPTH + 1) * sizeof(
                                             XmaFrame *));
        if (la_ctx->in_buf_Q == NULL) {
            XLNX_LA_LOG(XMA_ERROR_LOG, "OOM in_buf_Q\n");
            free(la_ctx);
            return NULL;
        }
    } else {
        la_ctx->in_buf_Q = NULL;
    }

    // Setup lookahead properties
    //filter_props = &la_ctx->filter_props;
    memset(&filter_props, 0, sizeof(XmaFilterProperties));
    filter_props.hwfilter_type = XMA_2D_FILTER_TYPE;
    strcpy(filter_props.hwvendor_string, "Xilinx");

    // Setup lookahead input port properties
    in_props = &filter_props.input;
    memset(in_props, 0, sizeof(XmaFilterPortProperties));
    in_props->format = cfg->fmt_type;
    in_props->bits_per_pixel = cfg->bits_per_pixel;
    in_props->width = cfg->width;
    in_props->height = cfg->height;
    in_props->stride = cfg->stride;
    in_props->framerate.numerator = cfg->framerate.numerator;
    in_props->framerate.denominator = cfg->framerate.denominator;

    // Setup lookahead output port properties
    out_props = &filter_props.output;
    memset(out_props, 0, sizeof(XmaFilterPortProperties));
    out_props->format = cfg->fmt_type; //XMA_VCU_NV12_FMT_TYPE;
    out_props->bits_per_pixel = cfg->bits_per_pixel;
    out_props->width = XLNX_ALIGN((in_props->width), 64)>>SCLEVEL1;
    out_props->height = XLNX_ALIGN((in_props->height), 64)>>SCLEVEL1;
    out_props->framerate.numerator = cfg->framerate.numerator;
    out_props->framerate.denominator = cfg->framerate.denominator;

    extn_params = &la_ctx->extn_params[0];
    extn_params[param_cnt].name = (char *)XLNX_LA_EXT_PARAMS[EParamIntraPeriod];
    extn_params[param_cnt].user_type = EParamIntraPeriod;
    extn_params[param_cnt].type = XMA_UINT32;
    extn_params[param_cnt].length = sizeof(int);
    extn_params[param_cnt].value = &cfg->gop_size;
    param_cnt++;

    extn_params[param_cnt].name = (char *)XLNX_LA_EXT_PARAMS[EParamLADepth];
    extn_params[param_cnt].user_type = EParamLADepth;
    extn_params[param_cnt].type = XMA_UINT32;
    extn_params[param_cnt].length = sizeof(int);
    extn_params[param_cnt].value = &la_ctx->lookahead_depth;
    param_cnt++;

    extn_params[param_cnt].name = (char *)XLNX_LA_EXT_PARAMS[EParamEnableZeroCopy];
    extn_params[param_cnt].user_type = EParamEnableZeroCopy;
    extn_params[param_cnt].type = XMA_UINT32;
    extn_params[param_cnt].length = sizeof(int);
    extn_params[param_cnt].value = &la_ctx->enableZeroCopy;
    param_cnt++;

    extn_params[param_cnt].name = (char *)XLNX_LA_EXT_PARAMS[EParamSpatialAQMode];
    extn_params[param_cnt].user_type = EParamSpatialAQMode;
    extn_params[param_cnt].type = XMA_UINT32;
    extn_params[param_cnt].length = sizeof(int);
    extn_params[param_cnt].value = &la_ctx->spatial_aq_mode;
    param_cnt++;

    extn_params[param_cnt].name = (char *)XLNX_LA_EXT_PARAMS[EParamTemporalAQMode];
    extn_params[param_cnt].user_type = EParamTemporalAQMode;
    extn_params[param_cnt].type = XMA_UINT32;
    extn_params[param_cnt].length = sizeof(int);
    extn_params[param_cnt].value = &la_ctx->temporal_aq_mode;
    param_cnt++;

    extn_params[param_cnt].name = (char *)XLNX_LA_EXT_PARAMS[EParamRateControlMode];
    extn_params[param_cnt].user_type = EParamRateControlMode;
    extn_params[param_cnt].type = XMA_UINT32;
    extn_params[param_cnt].length = sizeof(int);
    extn_params[param_cnt].value = &la_ctx->rate_control_mode;
    param_cnt++;

    extn_params[param_cnt].name = (char *)XLNX_LA_EXT_PARAMS[EParamSpatialAQGain];
    extn_params[param_cnt].user_type = EParamSpatialAQGain;
    extn_params[param_cnt].type = XMA_UINT32;
    extn_params[param_cnt].length = sizeof(int);
    extn_params[param_cnt].value = &la_ctx->spatial_aq_gain;
    param_cnt++;

    extn_params[param_cnt].name = (char *)XLNX_LA_EXT_PARAMS[EParamNumBFrames];
    extn_params[param_cnt].user_type = EParamNumBFrames;
    extn_params[param_cnt].type = XMA_UINT32;
    extn_params[param_cnt].length = sizeof(int);
    extn_params[param_cnt].value = &cfg->b_frames;
    param_cnt++;

    extn_params[param_cnt].name = (char *)XLNX_LA_EXT_PARAMS[EParamCodecType];
    extn_params[param_cnt].user_type = EParamCodecType;
    extn_params[param_cnt].type = XMA_UINT32;
    extn_params[param_cnt].length = sizeof(int);
    extn_params[param_cnt].value = &la_ctx->codec_type;
    param_cnt++;

    extn_params[param_cnt].name = (char *)XLNX_LA_EXT_PARAMS[EParamLatencyLogging];
    extn_params[param_cnt].user_type = EParamLatencyLogging;
    extn_params[param_cnt].type = XMA_UINT32;
    extn_params[param_cnt].length = sizeof(int);
    extn_params[param_cnt].value = &cfg->latency_logging;
    param_cnt++;

    filter_props.param_cnt = param_cnt;
    filter_props.params = &extn_params[0];

    /*----------------------------------------------------
      Allocate lookahead resource from XRM reserved resource
      ----------------------------------------------------*/
    la_ctx->lookahead_res_inuse = 0;
    _allocate_xrm_la_cu(la_ctx, &filter_props);

    // Create lookahead session based on the requested properties
    la_ctx->filter_session = xma_filter_session_create(&filter_props);
    if (!la_ctx->filter_session) {
        XLNX_LA_LOG(XMA_ERROR_LOG, "Failed to create lookahead session\n");
        destroy_xlnx_la(la_ctx);
        return NULL;
    }
    return (xlnx_lookahead_t)la_ctx;
}

int32_t destroy_xlnx_la(xlnx_lookahead_t la)
{
    xlnx_la_ctx *la_ctx;
    if (!la) {
        return XMA_ERROR;
    }
    la_ctx = (xlnx_la_ctx *)la;
    if (la_ctx->bypass == 0) {
        free_res(la_ctx);
    }
    free(la_ctx);
    return XMA_SUCCESS;
}

static int32_t xlnx_la_send_frame(xlnx_lookahead_t la, XmaFrame *in_frame)
{
    int32_t rc;
    xlnx_la_ctx *la_ctx = (xlnx_la_ctx *)la;
    if (!la_ctx) {
        XLNX_LA_LOG(XMA_ERROR_LOG, "xlnx_la_send_frame : XMA_ERROR\n");
        return XMA_ERROR;
    }

    if (in_frame && in_frame->do_not_encode) {
        rc = XMA_SUCCESS;
    } else {
        rc = xma_filter_session_send_frame(la_ctx->filter_session,
                                           in_frame);
    }
    if (rc <= XMA_ERROR) {
        XLNX_LA_LOG(XMA_ERROR_LOG,
                    "xlnx_la_send_frame : Send frame to LA xma plg Failed!!\n");
        rc = XMA_ERROR;
    }
    return rc;
}

static int32_t xlnx_la_recv_frame_with_qpmap(xlnx_lookahead_t la,
        XmaFrame *in_frame, XmaFrame **out_frame)
{
    int32_t rc;
    xlnx_la_ctx *la_ctx = (xlnx_la_ctx *)la;
    if (la_ctx->qp_frame == NULL) {
        if (la_ctx->in_buf_Q) {
            if(av_fifo_size(la_ctx->in_buf_Q) >= sizeof(XmaFrame *)) {
                av_fifo_generic_read(la_ctx->in_buf_Q, &la_ctx->qp_frame, sizeof(XmaFrame *),
                                     NULL);
            } else {
                return XMA_SEND_MORE_DATA;
            }
        } else {
            la_ctx->qp_frame = in_frame;
        }
    }

    if (la_ctx->qp_frame == NULL) {
        return XMA_SEND_MORE_DATA;
    }

    if (la_ctx->qp_frame->is_last_frame ||
            la_ctx->qp_frame->do_not_encode) {
        *out_frame = la_ctx->qp_frame;
        la_ctx->qp_frame = NULL;
        return XMA_SUCCESS;
    }
    rc = xma_filter_session_recv_frame(la_ctx->filter_session,
                                       la_ctx->qp_frame);
    if (rc == XMA_SUCCESS) {
        *out_frame = la_ctx->qp_frame;
        la_ctx->qp_frame = NULL;
    }

    return rc;
}

int32_t xlnx_la_send_recv_frame(xlnx_lookahead_t la, XmaFrame *in_frame,
                                XmaFrame **out_frame)
{
    int32_t ret = 0;
    xlnx_la_ctx *la_ctx = (xlnx_la_ctx *)la;
    if (la_ctx->bypass == 1) {
        *out_frame = in_frame;
        return XMA_SUCCESS;
    }
    if (*out_frame != NULL) {
        return XMA_ERROR;
    }
    ret = xlnx_la_send_frame(la, in_frame);
    switch (ret) {
        case XMA_TRY_AGAIN:
            ret = xlnx_la_recv_frame_with_qpmap(la, in_frame, out_frame);
            if (ret == XMA_SUCCESS) {
                ret = xlnx_la_send_frame(la, in_frame);
            }
            break;
        case XMA_SUCCESS:
            ret = xlnx_la_recv_frame_with_qpmap(la, in_frame, out_frame);
            if (ret == XMA_TRY_AGAIN) {
                ret = XMA_SEND_MORE_DATA;
            }
            break;
        case XMA_SEND_MORE_DATA:
            break;
        case XMA_ERROR:
        default:
            *out_frame = NULL;
            break;
    }
    if (ret == XMA_SEND_MORE_DATA || ret == XMA_SUCCESS) {
        if (in_frame && la_ctx->in_buf_Q) {
            av_fifo_generic_write(la_ctx->in_buf_Q, &in_frame, sizeof(XmaFrame *), NULL);
        }
    }
    return ret;
}
